#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
mn_entry.py - Mnemonic user entry methods for the MMGen suite
"""

import sys, time

from .util import msg, msg_r, fmt, fmt_list, capfirst, die, ascii_lowercase
from .term import get_char, get_char_raw
from .color import cyan

_return_chars = '\n\r '
_erase_chars = '\b\x7f'

class MnEntryMode:
	"""
	Subclasses must implement:
	  - pad_max:    pad character limit (None if variable)
	  - ss_len:     substring length for automatic entry
	  - get_word(): get a word from the user and return an index into the wordlist,
	                or None on failure
	"""

	pad_max_info = """
		Up to {pad_max}
		pad characters per word are permitted.
	"""

	def __init__(self, mne):
		self.pad_max_info = '  ' + self.pad_max_info.lstrip() if self.pad_max else '\n'
		self.mne = mne

	def get_char(self, s):
		did_erase = False
		while True:
			ch = get_char_raw('', num_bytes=1)
			if s and ch in _erase_chars:
				s = s[:-1]
				did_erase = True
			else:
				return (ch, s, did_erase)

class MnEntryModeFull(MnEntryMode):
	name = 'Full'
	choose_info = """
		Words must be typed in full and entered with ENTER, SPACE,
		or a pad character.
	"""
	prompt_info = """
		Use the ENTER or SPACE key to enter each word.  A pad character will also
		enter a word once you’ve typed {ssl} characters total (including pad chars).
	"""
	pad_max = None

	@property
	def ss_len(self):
		return self.mne.longest_word

	def get_word(self, mne):
		s, pad = ('', 0)
		while True:
			ch, s, _ = self.get_char(s)
			if ch in _return_chars:
				if s:
					break
			elif ch in ascii_lowercase:
				s += ch
			else:
				pad += 1
				if pad + len(s) > self.ss_len:
					break

		return mne.idx(s, 'full')

class MnEntryModeShort(MnEntryMode):
	name = 'Short'
	choose_info = """
		Words are entered automatically once {usl} valid word letters
		are typed.
	"""
	prompt_info = """
		Each word is entered automatically once {ssl} valid word letters are typed.
	"""
	prompt_info_bip39_add = """
		Words shorter than {ssl} letters can be entered with ENTER or SPACE, or by
		exceeding the pad character limit.
	"""
	pad_max = 16

	def __init__(self, mne):
		if mne.wl_id == 'bip39':
			self.prompt_info += '  ' + self.prompt_info_bip39_add.strip()
		super().__init__(mne)

	@property
	def ss_len(self):
		return self.mne.uniq_ss_len

	def get_word(self, mne):
		s, pad = ('', 0)
		while True:
			ch, s, _ = self.get_char(s)
			if ch in _return_chars:
				if s:
					break
			elif ch in ascii_lowercase:
				s += ch
				if len(s) == self.ss_len:
					break
			else:
				pad += 1
				if pad > self.pad_max:
					break

		return mne.idx(s, 'short')

class MnEntryModeFixed(MnEntryMode):
	name = 'Fixed'
	choose_info = """
		Words are entered automatically once exactly {usl} characters
		are typed.
	"""
	prompt_info = """
		Each word is entered automatically once exactly {ssl} characters are typed.
	"""
	prompt_info_add = ("""
		Words shorter than {ssl} letters must be padded to fit.
		""", """
		{sw}-letter words must be padded with one pad character.
		""")
	pad_max = None

	def __init__(self, mne):
		self.len_diff = mne.uniq_ss_len - mne.shortest_word
		self.prompt_info += self.prompt_info_add[self.len_diff==1].lstrip()
		super().__init__(mne)

	@property
	def ss_len(self):
		return self.mne.uniq_ss_len

	def get_word(self, mne):
		s, pad = ('', 0)
		while True:
			ch, s, _ = self.get_char(s)
			if ch in _return_chars:
				if s:
					break
			elif ch in ascii_lowercase:
				s += ch
				if len(s) + pad == self.ss_len:
					return mne.idx(s, 'short')
			else:
				pad += 1
				if pad > self.len_diff:
					return None
				if len(s) + pad == self.ss_len:
					return mne.idx(s, 'short')

class MnEntryModeMinimal(MnEntryMode):
	name = 'Minimal'
	choose_info = """
		Words are entered automatically once a minimum number of
		letters are typed (the number varies from word to word).
	"""
	prompt_info = """
		Each word is entered automatically once the minimum required number of valid
		word letters is typed.

		If your word is not entered automatically, that means it’s a substring of
		another word in the wordlist.  Such words must be entered explicitly with
		the ENTER or SPACE key, or by exceeding the pad character limit.
	"""
	pad_max = 16
	ss_len = None

	def get_word(self, mne):
		s, pad = ('', 0)
		lo, hi = (0, len(mne.wl) - 1)
		while True:
			ch, s, did_erase = self.get_char(s)
			if did_erase:
				lo, hi = (0, len(mne.wl) - 1)
			if ch in _return_chars:
				if s:
					return mne.idx(s, 'full', lo_idx=lo, hi_idx=hi)
			elif ch in ascii_lowercase:
				s += ch
				ret = mne.idx(s, 'minimal', lo_idx=lo, hi_idx=hi)
				if not isinstance(ret, tuple):
					return ret
				lo, hi = ret
			else:
				pad += 1
				if pad > self.pad_max:
					return mne.idx(s, 'full', lo_idx=lo, hi_idx=hi)

class MnemonicEntry:

	prompt_info = {
		'intro': """
			You will now be prompted for your {ml}-word seed phrase, one word at a time.
		""",
		'pad_info': """
			Note that anything you type that’s not a lowercase letter will simply be
			ignored.  This feature allows you to guard against acoustic side-channel
			attacks by padding your keyboard entry with “dead characters”.  Pad char-
			acters may be typed before, after, or in the middle of words.
		""",
	}
	word_prompt = ('Enter word #{}: ', 'Incorrect entry. Repeat word #{}: ')
	usr_dfl_entry_mode = None
	_lw = None
	_sw = None
	_usl = None

	def __init__(self, cfg):
		self.cfg = cfg
		self.set_dfl_entry_mode()

	@property
	def longest_word(self):
		if not self._lw:
			self._lw = max(len(w) for w in self.wl)
		return self._lw

	@property
	def shortest_word(self):
		if not self._sw:
			self._sw = min(len(w) for w in self.wl)
		return self._sw

	@property
	def uniq_ss_len(self):
		if not self._usl:
			usl = 0
			for i in range(len(self.wl)-1):
				w1, w2 = self.wl[i], self.wl[i+1]
				while True:
					if w1[:usl] == w2[:usl]:
						usl += 1
					else:
						break
			self._usl = usl
		return self._usl

	def idx(self, w, entry_mode, *, lo_idx=None, hi_idx=None):
		"""
		Return values:
		  - all modes:
		    - None:            failure (substr not in list)
		    - idx:             success
		  - minimal mode:
		    - (lo_idx,hi_idx): non-unique match
		"""
		trunc_len = {
			'full': self.longest_word,
			'short': self.uniq_ss_len,
			'minimal': len(w),
		}[entry_mode]
		w = w[:trunc_len]
		last_idx = len(self.wl) - 1
		lo = lo_idx or 0
		hi = hi_idx or last_idx
		while True:
			idx = (hi + lo) // 2
			cur_w = self.wl[idx][:trunc_len]
			if cur_w == w:
				if entry_mode == 'minimal':
					if idx > 0 and self.wl[idx-1][:len(w)] == w:
						return (lo, hi)
					elif idx < last_idx and self.wl[idx+1][:len(w)] == w:
						return (lo, hi)
				return idx
			elif hi <= lo:
				return None
			elif cur_w > w:
				hi = idx - 1
			else:
				lo = idx + 1

	def get_cls_by_entry_mode(self, entry_mode):
		return getattr(sys.modules[__name__], 'MnEntryMode' + capfirst(entry_mode))

	def choose_entry_mode(self):
		msg('Choose an entry mode:\n')
		em_objs = [self.get_cls_by_entry_mode(entry_mode)(self) for entry_mode in self.entry_modes]
		for n, mode in enumerate(em_objs, 1):
			msg('  {}) {:8} {}'.format(
				n,
				mode.name + ':',
				fmt(mode.choose_info, indent=' '*14).lstrip().format(usl=self.uniq_ss_len),
			))
		prompt = f'Type a number, or hit ENTER for the default ({capfirst(self.dfl_entry_mode)}): '
		erase = '\r' + ' ' * (len(prompt)+19) + '\r'
		while True:
			uret = get_char(prompt).strip()
			if uret == '':
				msg_r(erase)
				return self.get_cls_by_entry_mode(self.dfl_entry_mode)(self)
			elif uret in [str(i) for i in range(1, len(em_objs)+1)]:
				msg_r(erase)
				return em_objs[int(uret)-1]
			else:
				msg_r(f'\b {uret!r}: invalid choice ')
				time.sleep(self.cfg.err_disp_timeout)
				msg_r(erase)

	def get_mnemonic_from_user(self, mn_len, *, validate=True):
		mll = list(self.bconv.seedlen_map_rev)
		assert mn_len in mll, f'{mn_len}: invalid mnemonic length (must be one of {mll})'

		if self.usr_dfl_entry_mode:
			em = self.get_cls_by_entry_mode(self.usr_dfl_entry_mode)(self)
			i_add = ' (user-configured)'
		else:
			em = self.choose_entry_mode()
			i_add = ''

		msg('\r' + f'Using entry mode {cyan(em.name.upper())}{i_add}')
		self.em = em

		if not self.usr_dfl_entry_mode:
			msg('\n' + (
					fmt(self.prompt_info['intro'])
					+ '\n'
					+ fmt(self.prompt_info['pad_info'].rstrip() + em.pad_max_info + em.prompt_info, indent='  ')
				).format(
					ml       = mn_len,
					ssl      = em.ss_len,
					pad_max  = em.pad_max,
					sw       = self.shortest_word,
			))

		clear_line = '\n' if self.cfg.test_suite else '{r}{s}{r}'.format(r='\r', s=' '*40)
		idx, idxs = 1, [] # initialize idx to a non-None value

		while len(idxs) < mn_len:
			msg_r(self.word_prompt[idx is None].format(len(idxs)+1))
			idx = em.get_word(self)
			msg_r(clear_line)
			if idx is None:
				time.sleep(0.1)
			else:
				idxs.append(idx)

		words = [self.wl[i] for i in idxs]

		if validate:
			self.bconv.tohex(words)

		return ' '.join(words)

	@classmethod
	def get_cls_by_wordlist(cls, wl):
		d = {
			'mmgen': MnemonicEntryMMGen,
			'bip39': MnemonicEntryBIP39,
			'xmrseed': MnemonicEntryMonero,
		}
		wl = wl.lower()
		if wl not in d:
			raise ValueError(f'wordlist {wl!r} not recognized (valid choices: {fmt_list(list(d))})')
		return d[wl]

	def set_dfl_entry_mode(self):
		"""
		In addition to setting the default entry mode for the current wordlist, checks validity
		of all user-configured entry modes
		"""
		for k, v in self.cfg.mnemonic_entry_modes.items():
			cls = self.get_cls_by_wordlist(k)
			if v not in cls.entry_modes:
				errmsg = f"""
					Error in cfg file option 'mnemonic_entry_modes':
					Entry mode {v!r} not recognized for wordlist {k!r}:
					Valid choices: {fmt_list(cls.entry_modes)}
				"""
				die(2, '\n' + fmt(errmsg, indent='  '))
			if cls == type(self):
				self.usr_dfl_entry_mode = v

class MnemonicEntryMMGen(MnemonicEntry):
	wl_id = 'mmgen'
	modname = 'baseconv'
	entry_modes = ('full', 'minimal', 'fixed')
	dfl_entry_mode = 'minimal'
	has_chksum = False

class MnemonicEntryBIP39(MnemonicEntry):
	wl_id = 'bip39'
	modname = 'bip39'
	entry_modes = ('full', 'short', 'fixed')
	dfl_entry_mode = 'fixed'
	has_chksum = True

class MnemonicEntryMonero(MnemonicEntry):
	wl_id = 'xmrseed'
	modname = 'xmrseed'
	entry_modes = ('full', 'short')
	dfl_entry_mode = 'short'
	has_chksum = True

def mn_entry(cfg, wl_id, *, entry_mode=None):
	if wl_id == 'words':
		wl_id = 'mmgen'
	me = MnemonicEntry.get_cls_by_wordlist(wl_id)(cfg)
	import importlib
	me.bconv = getattr(importlib.import_module(f'mmgen.{me.modname}'), me.modname)(wl_id)
	me.wl = me.bconv.digits
	if entry_mode:
		me.em = getattr(sys.modules[__name__], 'MnEntryMode' + capfirst(entry_mode))(me)
	return me



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
derive: coin private key secret derivation for the MMGen suite
"""

from collections import namedtuple
from hashlib import sha512, sha256
from .addrlist import AddrIdxList

pk_bytes = namedtuple('coin_privkey_bytes', ['idx', 'pos', 'data'])

def derive_coin_privkey_bytes(seed, idxs):

	assert isinstance(idxs, AddrIdxList), f'{type(idxs)}: idx list not of type AddrIdxList'

	t_keys = len(idxs)
	pos = 0

	for idx in range(1, AddrIdxList.max_len + 1): # key/addr indexes begin from one

		seed = sha512(seed).digest()

		if idx == idxs[pos]:

			pos += 1

			# secret is double sha256 of seed hash round /idx/
			yield pk_bytes(idx, pos, sha256(sha256(seed).digest()).digest())

			if pos == t_keys:
				break



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
rpc.backends.curl: curl RPC backend for the MMGen Project
"""

import json

from ...util import ymsg

from ..util import dmsg_rpc, dmsg_rpc_backend, json_encoder

from .base import base

class curl(base):

	def __init__(self, caller):

		def gen_opts():
			for k, v in caller.http_hdrs.items():
				yield from ('--header', f'{k}: {v}')
			if caller.auth_type:
				# Authentication with curl is insecure, as it exposes the user's credentials
				# via the command line.  Use for testing only.
				yield from ('--user', f'{caller.auth.user}:{caller.auth.passwd}')
			if caller.auth_type == 'digest':
				yield '--digest'
			if caller.network_proto == 'https' and caller.verify_server is False:
				yield '--insecure'

		super().__init__(caller)
		self.exec_opts = list(gen_opts()) + ['--silent']
		self.arg_max = 8192 # set way below system ARG_MAX, just to be safe

	async def run(self, payload, timeout, host_path):
		data = json.dumps(payload, cls=json_encoder)
		if len(data) > self.arg_max:
			from .httplib import httplib
			ymsg('Warning: Curl data payload length exceeded - falling back on httplib')
			return httplib(self.caller).run(payload, timeout, host_path)
		dmsg_rpc_backend(self.host_url, host_path, payload)
		exec_cmd = [
			'curl',
			'--proxy', f'socks5h://{self.proxy}' if self.proxy else '',
			'--connect-timeout', str(timeout or self.timeout),
			'--write-out', '%{http_code}',
			'--data-binary', data
			] + self.exec_opts + [self.host_url + host_path]

		dmsg_rpc('    RPC curl exec data ==>\n{}\n', exec_cmd)

		from subprocess import run, PIPE
		from ...color import set_vt100
		res = run(exec_cmd, stdout=PIPE, check=True, text=True).stdout
		set_vt100()
		return (res[:-3], int(res[-3:]))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tool.wallet: Wallet routines for the 'mmgen-tool' utility
"""

from .common import tool_cmd_base

from ..subseed import SubSeedList
from ..seedsplit import MasterShareIdx
from ..wallet import Wallet

class tool_cmd(tool_cmd_base):
	"key, address or subseed generation from an MMGen wallet"

	def __init__(self, cfg, *, cmdname=None, proto=None, mmtype=None):
		self.need_proto = cmdname in ('gen_key', 'gen_addr')
		super().__init__(cfg, cmdname=cmdname, proto=proto, mmtype=mmtype)

	def _get_seed_file(self, wallet):
		from ..fileutil import get_seed_file
		return get_seed_file(
			cfg     = self.cfg,
			wallets = [wallet] if wallet else [],
			nargs   = 1)

	def get_subseed(self, subseed_idx: str, *, wallet=''):
		"get the Seed ID of a single subseed by Subseed Index for default or specified wallet"
		self.cfg._set_quiet(True)
		return Wallet(self.cfg, fn=self._get_seed_file(wallet)).seed.subseed(subseed_idx).sid

	def get_subseed_by_seed_id(self, seed_id: str, *, wallet='', last_idx=SubSeedList.dfl_len):
		"get the Subseed Index of a single subseed by Seed ID for default or specified wallet"
		self.cfg._set_quiet(True)
		ret = Wallet(
			self.cfg,
			fn = self._get_seed_file(wallet)).seed.subseed_by_seed_id(seed_id, last_idx=last_idx)
		return ret.ss_idx if ret else None

	def list_subseeds(self, subseed_idx_range: str, *, wallet=''):
		"list a range of subseed Seed IDs for default or specified wallet"
		self.cfg._set_quiet(True)
		from ..subseed import SubSeedIdxRange
		return Wallet(self.cfg, fn=self._get_seed_file(wallet)).seed.subseeds.format(
			*SubSeedIdxRange(subseed_idx_range))

	def list_shares(self,
			share_count: int,
			*,
			id_str = 'default',
			master_share: f'(min:1, max:{MasterShareIdx.max_val}, 0=no master share)' = 0,
			wallet = ''):
		"list the Seed IDs of the shares resulting from a split of default or specified wallet"
		self.cfg._set_quiet(True)
		return Wallet(self.cfg, fn=self._get_seed_file(wallet)).seed.split(
			share_count, id_str=id_str, master_idx=master_share).format()

	def gen_key(self, mmgen_addr: str, *, wallet=''):
		"generate a single WIF key for specified MMGen address from default or specified wallet"
		return self._gen_keyaddr(mmgen_addr, 'wif', wallet=wallet)

	def gen_addr(self, mmgen_addr: str, *, wallet=''):
		"generate a single MMGen address from default or specified wallet"
		return self._gen_keyaddr(mmgen_addr, 'addr', wallet=wallet)

	def _gen_keyaddr(self, mmgen_addr, target, *, wallet=''):
		from ..addr import MMGenID
		from ..addrlist import AddrList, AddrIdxList

		addr = MMGenID(self.proto, mmgen_addr)
		self.cfg._set_quiet(True)
		ss = Wallet(self.cfg, fn=self._get_seed_file(wallet))

		if ss.seed.sid != addr.sid:
			from ..util import die
			die(1, f'Seed ID of requested address ({addr.sid}) does not match wallet ({ss.seed.sid})')

		d = AddrList(
			cfg       = self.cfg,
			proto     = self.proto,
			seed      = ss.seed,
			addr_idxs = AddrIdxList(fmt_str=str(addr.idx)),
			mmtype    = addr.mmtype,
			skip_chksum = True).data[0]

		return {'wif': d.sec.wif, 'addr': d.addr}[target]



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.etc.params: Ethereum Classic protocol
"""

from ..eth.params import mainnet

class mainnet(mainnet):
	chain_names = ['classic', 'ethereum_classic']
	max_tx_fee  = 0.005
	coin_amt    = 'ETCAmt'

class testnet(mainnet):
	chain_names = ['morden', 'morden_testnet', 'classic-testnet']

class regtest(testnet):
	chain_names = ['developmentchain']
	decimal_prec = 64



#!/usr/bin/env python3
#
# Source URL:
# https://github.com/bitcoin/bitcoin/blob/v0.16.0/test/functional/test_framework/segwit_addr.py
#
# Copyright (c) 2017 Pieter Wuille
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.
#
# Unaltered except for the following changes by the MMGen Project:
#   leading spaces converted to tabs
#
"""Reference implementation for Bech32 and segwit addresses."""

CHARSET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"

def bech32_polymod(values):
	"""Internal function that computes the Bech32 checksum."""
	generator = [0x3b6a57b2, 0x26508e6d, 0x1ea119fa, 0x3d4233dd, 0x2a1462b3]
	chk = 1
	for value in values:
		top = chk >> 25
		chk = (chk & 0x1ffffff) << 5 ^ value
		for i in range(5):
			chk ^= generator[i] if ((top >> i) & 1) else 0
	return chk

def bech32_hrp_expand(hrp):
	"""Expand the HRP into values for checksum computation."""
	return [ord(x) >> 5 for x in hrp] + [0] + [ord(x) & 31 for x in hrp]

def bech32_verify_checksum(hrp, data):
	"""Verify a checksum given HRP and converted data characters."""
	return bech32_polymod(bech32_hrp_expand(hrp) + data) == 1

def bech32_create_checksum(hrp, data):
	"""Compute the checksum values given HRP and data."""
	values = bech32_hrp_expand(hrp) + data
	polymod = bech32_polymod(values + [0, 0, 0, 0, 0, 0]) ^ 1
	return [(polymod >> 5 * (5 - i)) & 31 for i in range(6)]

def bech32_encode(hrp, data):
	"""Compute a Bech32 string given HRP and data values."""
	combined = data + bech32_create_checksum(hrp, data)
	return hrp + '1' + ''.join([CHARSET[d] for d in combined])

def bech32_decode(bech):
	"""Validate a Bech32 string, and determine HRP and data."""
	if ((any(ord(x) < 33 or ord(x) > 126 for x in bech)) or
			(bech.lower() != bech and bech.upper() != bech)):
		return (None, None)
	bech = bech.lower()
	pos = bech.rfind('1')
	if pos < 1 or pos + 7 > len(bech) or len(bech) > 90:
		return (None, None)
	if not all(x in CHARSET for x in bech[pos+1:]):
		return (None, None)
	hrp = bech[:pos]
	data = [CHARSET.find(x) for x in bech[pos+1:]]
	if not bech32_verify_checksum(hrp, data):
		return (None, None)
	return (hrp, data[:-6])

def convertbits(data, frombits, tobits, pad=True):
	"""General power-of-2 base conversion."""
	acc = 0
	bits = 0
	ret = []
	maxv = (1 << tobits) - 1
	max_acc = (1 << (frombits + tobits - 1)) - 1
	for value in data:
		if value < 0 or (value >> frombits):
			return None
		acc = ((acc << frombits) | value) & max_acc
		bits += frombits
		while bits >= tobits:
			bits -= tobits
			ret.append((acc >> bits) & maxv)
	if pad:
		if bits:
			ret.append((acc << (tobits - bits)) & maxv)
	elif bits >= frombits or ((acc << (tobits - bits)) & maxv):
		return None
	return ret

def decode(hrp, addr):
	"""Decode a segwit address."""
	hrpgot, data = bech32_decode(addr)
	if hrpgot != hrp:
		return (None, None)
	decoded = convertbits(data[1:], 5, 8, False)
	if decoded is None or len(decoded) < 2 or len(decoded) > 40:
		return (None, None)
	if data[0] > 16:
		return (None, None)
	if data[0] == 0 and len(decoded) != 20 and len(decoded) != 32:
		return (None, None)
	return (data[0], decoded)

def encode(hrp, witver, witprog):
	"""Encode a segwit address."""
	ret = bech32_encode(hrp, [witver] + convertbits(witprog, 8, 5))
	if decode(hrp, ret) == (None, None):
		return None
	return ret



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
xmrseed: Monero mnemonic conversion class for the MMGen suite
"""

from .baseconv import baseconv
from .util import die

def is_xmrseed(s):
	return bool(xmrseed().tobytes(s.split()))

# implements a subset of the baseconv API
class xmrseed(baseconv):

	desc            = baseconv.dt('Monero mnemonic', 'Monero new-style mnemonic seed phrase')
	wl_chksum       = '3c381ebb'
	seedlen_map     = {32: 25}
	seedlen_map_rev = {25: 32}

	def __init__(self, wl_id='xmrseed'):
		assert wl_id == 'xmrseed', "initialize with 'xmrseed' for compatibility with baseconv API"
		from .wordlist.monero import words
		self.digits = words
		self.wl_id = 'xmrseed'

	@staticmethod
	def monero_mn_checksum(words):
		from binascii import crc32
		wstr = ''.join(word[:3] for word in words)
		return words[crc32(wstr.encode()) % len(words)]

	def tobytes(self, words_arg, *, pad=None):

		assert isinstance(words_arg, (list, tuple)), 'words must be list or tuple'
		assert pad is None, f"{pad}: invalid 'pad' argument (must be None)"

		words = words_arg

		desc = self.desc.short
		wl = self.digits
		base = len(wl)

		if not set(words) <= set(wl):
			die('MnemonicError',  f'{words!r}: not in {desc} format')

		if len(words) not in self.seedlen_map_rev:
			die('MnemonicError',  f'{len(words)}: invalid seed phrase length for {desc}')

		z = self.monero_mn_checksum(words[:-1])
		if z != words[-1]:
			die('MnemonicError', f'invalid {desc} checksum')

		words = tuple(words[:-1])

		def gen():
			for i in range(len(words)//3):
				w1, w2, w3 = [wl.index(w) for w in words[3*i:3*i+3]]
				x = w1 + base*((w2-w1)%base) + base*base*((w3-w2)%base)
				yield x.to_bytes(4, 'big')[::-1]

		return b''.join(gen())

	def frombytes(self, bytestr, *, pad=None, tostr=False):
		assert pad is None, f"{pad}: invalid 'pad' argument (must be None)"

		desc = self.desc.short
		wl = self.digits
		base = len(wl)

		if len(bytestr) not in self.seedlen_map:
			die('SeedLengthError', f'{len(bytestr)}: invalid seed byte length for {desc}')

		def num2base_monero(num):
			w1 = num % base
			w2 = (num//base + w1) % base
			w3 = (num//base//base + w2) % base
			return (wl[w1], wl[w2], wl[w3])

		def gen():
			for i in range(len(bytestr)//4):
				yield from num2base_monero(int.from_bytes(bytestr[i*4:i*4+4][::-1], 'big'))

		o = list(gen())
		o.append(self.monero_mn_checksum(o))

		return ' '.join(o) if tostr else tuple(o)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.rune.tx.new: THORChain new transaction class
"""

from ....tx import new as TxBase

from ...vm.tx.new import New as VmNew

from .base import Base

class New(VmNew, Base, TxBase.New):

	async def get_fee(self, fee, outputs_sum, start_fee_desc):
		return await self.twctl.get_balance(self.inputs[0].addr)

	async def set_gas(self, *, to_addr=None, force=False):
		self.gas = self.dfl_gas

	def set_gas_with_data(self, data):
		pass

	def update_txid(self):
		return super().update_txid(
			self.serialized |
			({'memo': self.swap_memo} if self.is_swap else {}))

	async def make_txobj(self): # called by create_serialized()
		acct_info = self.rpc.get_account_info(self.inputs[0].addr)
		self.txobj = {
			'from':           self.inputs[0].addr,
			'to':             self.outputs[0].addr if self.outputs else None,
			'amt':            self.outputs[0].amt if self.outputs else self.swap_amt,
			'gas':            self.gas,
			'account_number': int(acct_info['account_number']),
			'sequence':       int(acct_info['sequence']),
			'chain_id':       self.proto.chain_id}



#!/usr/bin/env python3

"""
test.modtest_d.flags: unit test for the MMGen suite's ClassFlags class
"""

from mmgen.flags import ClassOpts, ClassFlags

from ..include.common import qmsg, qmsg_r, vmsg

class unit_test:

	def run_test(self, name, ut):

		class MyClassOpts(ClassOpts):
			reserved_attrs = ('foo',)

		class cls1:
			avail_opts = ()
			avail_flags = ()
			def __init__(self, opts=None, flags=None):
				self.opt = ClassOpts(self, opts)
				self.flag = ClassFlags(self, flags)

		class cls2(cls1):
			avail_opts = ('foo', 'bar')
			avail_flags = ('baz',)

		class cls3(cls1):
			avail_opts = ('_foo',)

		class cls4(cls1):
			avail_opts = ('foo',)
			def __init__(self, opts=None, flags=None):
				self.opt = MyClassOpts(self, opts)

		def test_flags():
			def gen():
				for n, cls in enumerate((
						cls1(),
						cls2(),
						cls2(opts=['bar']),
						cls2(flags=['baz']),
					)):
					vmsg(f'Cfg {n+1}:')
					for k in ('opt', 'flag'):
						vmsg(f'  {k}s: {getattr(cls, k)}')
					yield cls
			return list(gen())

		def test_flags_err(d):

			def bad1(): d[1].flag.foo = False
			def bad2(): d[1].opt.baz = False
			def bad3(): cls3()
			def bad4(): cls4()
			def bad5(): cls1(opts='foo')
			def bad6(): cls2(opts=['qux'])
			def bad7(): d[1].flag.baz = False
			def bad8(): d[3].flag.baz = True
			def bad9(): d[1].flag.baz = 'x'

			ut.process_bad_data((
				('flag (1)',             'ClassFlagsError', 'unrecognized flag', bad1),
				('opt (1)',              'ClassFlagsError', 'unrecognized opt',  bad2),
				('avail_opts (1)',       'ClassFlagsError', 'underscore',        bad3),
				('avail_opts (2)',       'ClassFlagsError', 'reserved name',     bad4),
				('class invocation (1)', 'AssertionError',  'list or tuple',     bad5),
				('class invocation (2)', 'ClassFlagsError', 'unrecognized opt',  bad6),
				('flag (2)',             'ClassFlagsError', 'not set',           bad7),
				('flag (3)',             'ClassFlagsError', 'already set',       bad8),
				('flag (4)',             'AssertionError',  'not boolean',       bad9),
			))

		qmsg_r('Testing flags and opts...')
		vmsg('')
		classes = test_flags()
		qmsg('OK')

		qmsg_r('Testing error handling for flags and opts...')
		vmsg('')
		test_flags_err(classes)
		qmsg('OK')

		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.sweep: Monero wallet ops for the MMGen Suite
"""

from ...util import msg, msg_r, gmsg, die, fmt_dict, make_timestr
from ...proto.xmr.rpc import MoneroWalletRPCClient
from ...proto.xmr.daemon import MoneroWalletDaemon
from ...ui import keypress_confirm

from .. import tx_priorities
from ..rpc import MoneroWalletRPC

from .spec import OpMixinSpec
from .wallet import OpWallet

class OpSweep(OpMixinSpec, OpWallet):
	spec_id  = 'sweep_spec'
	spec_key = ((1, 'source'), (3, 'dest'))
	opts = (
		'no_relay',
		'tx_relay_daemon',
		'watch_only',
		'priority',
		'skip_empty_accounts',
		'skip_empty_addresses')
	sweep_type = 'single-address'

	def check_uopts(self):
		if self.cfg.tx_relay_daemon and (self.cfg.no_relay or self.cfg.autosign):
			die(1, '--tx-relay-daemon makes no sense in this context!')

		if self.cfg.priority and self.cfg.priority not in list(tx_priorities):
			die(1, '{}: invalid parameter for --priority (valid params: {})'.format(
				self.cfg.priority,
				fmt_dict(tx_priorities, fmt='square_compact')))

	def init_tx_relay_daemon(self):

		m = self.parse_tx_relay_opt()

		wd2 = MoneroWalletDaemon(
			cfg         = self.cfg,
			proto       = self.proto,
			wallet_dir  = self.cfg.wallet_dir or '.',
			test_suite  = self.cfg.test_suite,
			monerod_addr = m[1],
			proxy       = m[2])

		if self.cfg.test_suite:
			wd2.usr_daemon_args = ['--daemon-ssl-allow-any-cert']

		wd2.start()

		self.c = MoneroWalletRPCClient(
			cfg    = self.cfg,
			daemon = wd2)

	def create_tx(self, h, wallet_data):

		def create_new_addr_maybe(h, account, label):
			if keypress_confirm(self.cfg, f'\nCreate new address for account #{account}?'):
				return h.create_new_addr(account, label)
			else:
				keypress_confirm(
					self.cfg,
					f'Sweep to last existing address of account #{account}?',
					do_exit = True)
				return None

		dest_addr_chk = None

		if self.dest is None: # sweep to same account
			dest_acct = self.account
			dest_addr_chk = create_new_addr_maybe(
				h, self.account, f'{self.name} from this account [{make_timestr()}]')
			if dest_addr_chk:
				wallet_data = h.get_wallet_data(print=False)
			dest_addr, dest_addr_idx = h.get_last_addr(self.account, wallet_data, display=not dest_addr_chk)
			if dest_addr_chk:
				h.print_acct_addrs(wallet_data, self.account)
		elif self.dest_acct is None: # sweep to wallet
			h.close_wallet('source')
			h2 = MoneroWalletRPC(self, self.dest)
			h2.open_wallet('destination')
			wallet_data2 = h2.get_wallet_data()

			wf = self.get_wallet_fn(self.dest)
			if keypress_confirm(self.cfg, f'\nCreate new account for wallet {wf.name!r}?'):
				dest_acct, dest_addr = h2.create_acct(
					label = f'{self.name} from {self.source.idx}:{self.account} [{make_timestr()}]')
				dest_addr_idx = 0
				h2.get_wallet_data()
			else:
				keypress_confirm(
					self.cfg,
					f'Sweep to last existing account of wallet {wf.name!r}?',
					do_exit = True)
				dest_acct, dest_addr_chk = h2.get_last_acct(wallet_data2.accts_data)
				dest_addr, dest_addr_idx = h2.get_last_addr(dest_acct, wallet_data2, display=False)

			h2.close_wallet('destination')
			h.open_wallet('source', refresh=False)
		else: # sweep to specific account of wallet

			def get_dest_addr_params(h, wallet_data, dest_acct, label):
				self.check_account_exists(wallet_data.accts_data, dest_acct)
				h.print_acct_addrs(wallet_data, dest_acct)
				dest_addr_chk = create_new_addr_maybe(h, dest_acct, label)
				if dest_addr_chk:
					wallet_data = h.get_wallet_data(print=False)
				dest_addr, dest_addr_idx = h.get_last_addr(dest_acct, wallet_data, display=not dest_addr_chk)
				if dest_addr_chk:
					h.print_acct_addrs(wallet_data, dest_acct)
				return dest_addr, dest_addr_idx, dest_addr_chk

			dest_acct = self.dest_acct

			if self.dest == self.source:
				dest_addr, dest_addr_idx, dest_addr_chk = get_dest_addr_params(
					h, wallet_data, dest_acct,
					f'{self.name} from account #{self.account} [{make_timestr()}]')
			else:
				h.close_wallet('source')
				h2 = MoneroWalletRPC(self, self.dest)
				h2.open_wallet('destination')
				dest_addr, dest_addr_idx, dest_addr_chk = get_dest_addr_params(
					h2, h2.get_wallet_data(), dest_acct,
					f'{self.name} from {self.source.idx}:{self.account} [{make_timestr()}]')
				h2.close_wallet('destination')
				h.open_wallet('source', refresh=False)

		assert dest_addr_chk in (None, dest_addr), (
			f'dest_addr: ({dest_addr}) != dest_addr_chk: ({dest_addr_chk})')

		msg(f'\n    Creating {self.name} transaction...')
		return (h, h.make_sweep_tx(self.account, dest_acct, dest_addr_idx, dest_addr, wallet_data.addrs_data))

	@property
	def add_desc(self):
		return (
			r' to new address' if self.dest is None else
			f' to new account in wallet {self.dest.idx}' if self.dest_acct is None else
			f' to account #{self.dest_acct} of wallet {self.dest.idx}') + f' ({self.sweep_type} sweep)'

	def check_account_exists(self, accts_data, idx):
		max_acct = len(accts_data['subaddress_accounts']) - 1
		if self.account > max_acct:
			die(2, f'{self.account}: requested account index out of bounds (>{max_acct})')

	async def main(self):

		gmsg(
			f'\n{self.stem.capitalize()}ing account #{self.account}'
			f' of wallet {self.source.idx}{self.add_desc}')

		h = MoneroWalletRPC(self, self.source)

		h.open_wallet('source')

		wallet_data = h.get_wallet_data(skip_empty_ok=True)

		self.check_account_exists(wallet_data.accts_data, self.account)

		h.print_acct_addrs(wallet_data, self.account)

		h, new_tx = self.create_tx(h, wallet_data)

		msg('\n' + new_tx.get_info(indent='    '))

		if self.cfg.tx_relay_daemon:
			self.display_tx_relay_info(indent='    ')

		msg('Saving TX data to file')
		new_tx.write(delete_metadata=True)

		if self.cfg.no_relay or self.cfg.autosign:
			return True

		keypress_confirm(
			self.cfg,
			f'Relay {self.name} transaction?',
			do_exit = True,
			exit_msg = '\nExiting at user request')

		if self.cfg.tx_relay_daemon:
			await h.stop_wallet('source')
			msg('')
			self.init_tx_relay_daemon()
			h = MoneroWalletRPC(self, self.source)
			h.open_wallet('TX-relay-configured source', refresh=False)
		msg_r(f'\n    Relaying {self.name} transaction...')
		h.relay_tx(new_tx.data.metadata)
		gmsg('\nAll done')
		return True

class OpSweepAll(OpSweep):
	stem = 'sweep'
	sweep_type = 'all-address'

class OpTransfer(OpSweep):
	stem    = 'transferr'
	spec_id = 'transfer_spec'
	spec_key = ((1, 'source'),)

	@property
	def add_desc(self):
		return f': {self.amount} XMR to {self.dest_addr}'

	def create_tx(self, h, wallet_data):
		msg(f'\n    Creating {self.name} transaction...')
		return (h, h.make_transfer_tx(self.account, self.dest_addr, self.amount))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.cmdtest_d.help: helpscreen test group for the cmdtest.py test suite
"""

import sys, os, time

from mmgen.util import ymsg
from mmgen.cfg import gc

from .base import CmdTestBase

class CmdTestHelp(CmdTestBase):
	'help, info and usage screens'
	networks = ('btc', 'ltc', 'bch', 'eth', 'xmr', 'doge', 'rune')
	passthru_opts = ('daemon_data_dir', 'rpc_port', 'coin', 'testnet')
	cmd_group = (
		('usage1',            (1, 'usage message (via --usage)', [])),
		('usage2',            (1, 'usage message (via --usage, with --coin)', [])),
		('usage3',            (1, 'usage message (via bad invocation)', [])),
		('usage4',            (1, 'usage message (via bad invocation, with --coin)', [])),
		('version',           (1, 'version message', [])),
		('license',           (1, 'license message', [])),
		('helpscreens',       (1, 'help screens',             [])),
		('longhelpscreens',   (1, 'help screens (--longhelp)', [])),
		('show_hash_presets', (1, 'info screen (--show-hash-presets)', [])),
		('tool_help',         (1, '‘mmgen-tool’ usage screen', [])),
		('tool_cmd_usage',    (1, '‘mmgen-tool’ usage screen', [])),
		('test_help',         (1, '‘cmdtest.py’ help screens', [])),
		('tooltest_help',     (1, '‘tooltest.py’ help screens', [])),
	)

	def usage1(self):
		return self._usage('walletgen', ['--usage'], True, False, 0)

	def usage2(self):
		return self._usage('tool' if self.coin == 'xmr' else 'txcreate', ['--usage'], True, True, 0)

	def usage3(self):
		return self._usage('walletgen', ['foo'], True, False, 1)

	def usage4(self):
		return self._usage('tool' if self.coin == 'xmr' else 'txcreate', [], True, True, 1)

	def _usage(self, cmd_arg, args, no_passthru_opts, add_coin_opt, exit_val):
		if cmd := (None if self._gen_skiplist(cmd_arg) else cmd_arg):
			t = self.spawn(
				f'mmgen-{cmd}',
				([f'--coin={self.coin}'] if add_coin_opt else []) + args,
				exit_val = exit_val,
				no_passthru_opts = no_passthru_opts)
			t.expect(f'USAGE: mmgen-{cmd}')
			return t
		return 'skip'

	def version(self):
		t = self.spawn('mmgen-tool', ['--version'], exit_val=0)
		t.expect('MMGEN-TOOL version')
		return t

	def license(self):
		t = self.spawn(
			'mmgen-walletconv',
			['--stdout', '--in-fmt=hex', '--out-fmt=hex'],
			env = {'MMGEN_NO_LICENSE':''},
			no_passthru_opts = True)
		t.expect('to continue: ', 'w')
		t.expect('TERMS AND CONDITIONS') # start of GPL text
		if self.cfg.pexpect_spawn:
			t.send('G')
		t.expect('return for a fee.')    # end of GPL text
		if self.cfg.pexpect_spawn:
			t.send('q')
		t.expect('to continue: ', 'c')
		t.expect('data: ', 'beadcafe'*4 + '\n')
		t.expect('to confirm: ', 'YES\n')
		return t

	def spawn_chk_expect(self, *args, **kwargs):
		expect = kwargs.pop('expect')
		t = self.spawn(*args, **kwargs)
		t.expect(expect)
		if t.pexpect_spawn:
			time.sleep(0.4)
			t.send('q')
		t.read()
		t.ok()
		t.skip_ok = True
		return t

	def _gen_skiplist(self, scripts):
		def gen(scripts):
			if isinstance(scripts, str):
				scripts = [scripts]
			for script in scripts:
				d = gc.cmd_caps_data[script]
				if sys.platform == 'win32' and 'w' not in d.platforms:
					yield script
				elif not (d.use_coin_opt or self.proto.coin.lower() == 'btc'):
					yield script
				else:
					for cap in d.caps:
						if cap not in self.proto.mmcaps:
							yield script
							break
		return set(gen(scripts))

	def helpscreens(self, arg='--help', scripts=(), expect='USAGE:.*OPTIONS:', pager=True):

		scripts = list(scripts or gc.cmd_caps_data)

		cmdlist = sorted(
			set(scripts)
			- self._gen_skiplist(scripts)
			- (set(gc.altcoin_cmds if self.cfg.no_altcoin else [])))

		for cmdname in cmdlist:
			cmd_caps = gc.cmd_caps_data[cmdname]
			assert cmd_caps, cmdname
			t = self.spawn(
				f'mmgen-{cmdname}',
				[arg],
				extra_desc       = f'(mmgen-{cmdname})',
				no_passthru_opts = not cmd_caps.use_coin_opt)
			t.expect(expect, regex=True)
			if pager and t.pexpect_spawn:
				time.sleep(0.2)
				t.send('q')
			t.read()
			t.ok()
			t.skip_ok = True

		return 'silent'

	def longhelpscreens(self):
		return self.helpscreens(arg='--longhelp', expect='USAGE:.*GLOBAL OPTIONS:')

	def show_hash_presets(self):
		return self.helpscreens(
			arg = '--show-hash-presets',
			scripts = (
				'walletgen', 'walletconv', 'walletchk', 'passchg', 'subwalletgen',
				'addrgen', 'keygen', 'passgen',
				'txdo', 'swaptxdo', 'txsign', 'txbump'),
			expect = 'Available parameters.*Preset',
			pager  = False)

	def tool_help(self):

		if os.getenv('PYTHONOPTIMIZE') == '2':
			ymsg('Skipping tool help with PYTHONOPTIMIZE=2 (no docstrings)')
			return 'skip'

		for arg in (
			'help',
			'usage',
		):
			t = self.spawn_chk_expect(
				'mmgen-tool',
				[arg],
				extra_desc = f'(mmgen-tool {arg})',
				expect = 'GENERAL USAGE')
		return t

	def tool_cmd_usage(self):

		if os.getenv('PYTHONOPTIMIZE') == '2':
			ymsg('Skipping tool cmd usage with PYTHONOPTIMIZE=2 (no docstrings)')
			return 'skip'

		from mmgen.main_tool import mods

		for cmdlist in mods.values():
			for cmd in cmdlist:
				t = self.spawn_chk('mmgen-tool', ['help', cmd], extra_desc=f'({cmd})')
		return t

	def test_help(self):
		for arg, expect in (
			('--help', 'USAGE'),
			('--list-cmds', 'AVAILABLE COMMANDS'),
			('--list-cmd-groups', 'AVAILABLE COMMAND GROUPS')
		):
			t = self.spawn_chk_expect(
				'cmdtest.py',
				[arg],
				cmd_dir = 'test',
				extra_desc = f'(cmdtest.py {arg})',
				expect = expect)
		return t

	def tooltest_help(self):
		for arg, expect in (
			('--list-cmds', 'Available commands'),
			('--testing-status', 'Testing status')
		):
			t = self.spawn_chk_expect(
				'tooltest.py',
				[arg],
				cmd_dir = 'test',
				extra_desc = f'(tooltest.py {arg})',
				expect = expect)
		return t



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.ltc.params: Litecoin protocol
"""

from ..btc.params import mainnet

class mainnet(mainnet):
	block0          = '12a765e31ffd4059bada1e25190f6e98c99d9714d334efa41a195a7e7e04bfe2'
	addr_ver_info   = {'30': 'p2pkh', '05': 'p2sh', '32': 'p2sh'} # new p2sh ver 0x32 must come last
	wif_ver_num     = {'std': 'b0'}
	mmtypes         = ('L', 'C', 'S', 'B')
	coin_amt        = 'LTCAmt'
	max_tx_fee      = 0.3
	base_coin       = 'LTC'
	forks           = []
	bech32_hrp      = 'ltc'
	avg_bdi         = 150
	halving_interval = 840000

class testnet(mainnet):
	# addr ver nums same as Bitcoin testnet, except for 'p2sh'
	addr_ver_info      = {'6f': 'p2pkh', 'c4': 'p2sh', '3a': 'p2sh'}
	wif_ver_num        = {'std': 'ef'} # same as Bitcoin testnet
	bech32_hrp         = 'tltc'

class regtest(testnet):
	bech32_hrp         = 'rltc'
	halving_interval   = 150



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
help: help notes functions for MMGen suite commands
"""

class help_notes:

	def __init__(self, proto, cfg):
		self.proto = proto
		self.cfg = cfg

	def txcreate_args(self):
		return (
			'[ADDR,AMT ... | DATA_SPEC] ADDR'
				if self.proto.base_proto == 'Bitcoin' else
			'ADDR,AMT')

	def swaptxcreate_args(self):
		return 'COIN1 [AMT CHG_ADDR] COIN2 [ADDR]'

	def account_info_desc(self):
		return 'unspent outputs' if self.proto.base_proto == 'Bitcoin' else 'account info'

	def fee_spec_letters(self, *, use_quotes=False, proto=None):
		cu = (proto or self.proto).coin_amt.units
		pfx, sfx, sep, conj = (('', '', ',', ' or '), ("‘", "’", "’,‘", "’ or ‘"))[use_quotes]
		return pfx + sep.join(u[0] for u in cu[:-1]) + ('', conj)[len(cu)>1] + cu[-1][0] + sfx

	def stream_interval(self):
		from ..tx.new_swap import get_swap_proto_mod
		return get_swap_proto_mod(self.cfg.swap_proto).SwapCfg(self.cfg).si.dfl

	def fee_spec_names(self, *, proto=None, linebreak=' '):
		cu = (proto or self.proto).coin_amt.units
		return (
			', '.join(cu[:-1])
			+ ('', ' and ')[len(cu)>1]
			+ cu[-1]
			+ (f',{linebreak}respectively' if len(cu) > 1 else ''))

	def dfl_twname(self):
		from ..proto.btc.rpc.local import BitcoinRPCClient
		return BitcoinRPCClient.dfl_twname

	def MasterShareIdx(self):
		from ..seedsplit import MasterShareIdx
		return MasterShareIdx

	def tool_help(self):
		from ..tool.help import main_help
		return main_help()

	def dfl_subseeds(self):
		from ..subseed import SubSeedList
		return str(SubSeedList.dfl_len)

	def dfl_seed_len(self):
		from ..seed import Seed
		return str(Seed.dfl_len)

	def password_formats(self):
		from ..passwdlist import PasswordList
		pwi_fs = '{:8} {:1} {:26} {:<7}  {:<7}  {}'
		return '\n  '.join(
			[pwi_fs.format('Code', '', 'Description', 'Min Len', 'Max Len', 'Default Len')] +
			[pwi_fs.format(k, '-', v.desc, v.min_len, v.max_len, v.dfl_len)
				for k, v in PasswordList.pw_info.items()]
		)

	def dfl_mmtype(self):
		from ..addr import MMGenAddrType
		return "'{}' or '{}'".format(self.proto.dfl_mmtype, MMGenAddrType.mmtypes[self.proto.dfl_mmtype].name)

	def address_types(self):
		from ..addr import MMGenAddrType
		return """
ADDRESS TYPES:

  Code Type           Description
  ---- ----           -----------
  """ + format('\n  '.join(['‘{}’  {:<12} - {}'.format(k, v.name, v.desc)
				for k, v in MMGenAddrType.mmtypes.items()]))

	def fmt_codes(self):
		from ..wallet import format_fmt_codes
		return """
FMT CODES:

  """ + '\n  '.join(format_fmt_codes().splitlines())

	def coin_id(self):
		return self.proto.coin_id

	def keygen_backends(self):
		from ..keygen import get_backends
		from ..addr import MMGenAddrType
		backends = get_backends(
			MMGenAddrType(self.proto, self.cfg.type or self.proto.dfl_mmtype).pubkey_type
		)
		return ' '.join('{n}:{k}{t}'.format(n=n, k=k, t=('', ' [default]')[n == 1])
			for n, k in enumerate(backends, 1))

	def coin_daemon_network_ids(self):
		from ..daemon import CoinDaemon
		from ..util import fmt_list
		return fmt_list(CoinDaemon.get_network_ids(self.cfg), fmt='bare')

	def tx_proxies(self):
		from ..util import fmt_list
		return fmt_list(self.cfg._autoset_opts['tx_proxy'].choices, fmt='fancy')

	def rel_fee_desc(self):
		if self.proto.has_usr_fee:
			from ..tx import BaseTX
			return BaseTX(cfg=self.cfg, proto=self.proto).rel_fee_desc
		else:
			return ''

	def gas_limit(self, target):
		return """
                                 GAS LIMIT

This option specifies the maximum gas allowance for an Ethereum transaction.
It’s generally of interest only for token transactions or swap transactions
from token assets.

Parameter must be an integer or one of the special values ‘fallback’ (for a
locally computed sane default) or ‘auto’ (for gas estimate via an RPC call,
in the case of a token transaction, or locally computed default, in the case
of a standard transaction). The default is ‘auto’.

		""" if target == 'swaptx' or self.proto.base_coin == 'ETH' else ''

	def fee(self, all_coins=False):

		if not self.proto.has_usr_fee:
			return ''

		from ..tx import BaseTX
		text = """
                               FEE SPECIFICATION

Transaction fees, both on the command line and at the interactive prompt, may
be specified as either absolute coin amounts, using a plain decimal number, or
as {r}, using an integer followed by {l}, for{s}{u}""".format(
			r = BaseTX(cfg=self.cfg, proto=self.proto).rel_fee_desc,
			l = self.fee_spec_letters(use_quotes=True),
			s = '\n' if self.proto.base_coin == 'ETH' else ' ',
			u = self.fee_spec_names())

		if all_coins:
			from ..protocol import init_proto
			eth_proto = init_proto(self.cfg, 'eth', need_amt=True)
			return text + (
				' (for\nBitcoin, Litecoin and Bitcoin Cash)'
				+ ", or {r}, using an integer followed\nby {l}, for {u}".format(
					r = BaseTX(cfg=self.cfg, proto=eth_proto).rel_fee_desc,
					l = self.fee_spec_letters(use_quotes=True, proto=eth_proto),
					u = self.fee_spec_names(proto=eth_proto, linebreak='\n'))
				+ ' (for Ethereum)\n\n')
		else:
			return text + '.\n\n'

	def passwd(self):
		return """
PASSPHRASE NOTE:

For passphrases all combinations of whitespace are equal, and leading and
trailing space are ignored.  This permits reading passphrase or brainwallet
data from a multi-line file with free spacing and indentation.
""".strip()

	def brainwallet(self):
		return """
BRAINWALLET NOTE:

To thwart dictionary attacks, it’s recommended to use a strong hash preset
with brainwallets.  For a brainwallet passphrase to generate the correct
seed, the same seed length and hash preset parameters must always be used.
""".strip()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
proto.eth.tw.bal: Ethereum tracking wallet getbalance class
"""

from ....tw.ctl import TwCtl
from ....tw.bal import TwGetBalance

class EthereumTwGetBalance(TwGetBalance):

	start_labels = ('TOTAL', 'Non-MMGen')
	conf_cols = {
		'ge_minconf': 'Balance',
	}

	async def __init__(self, cfg, proto, *, minconf, quiet):
		self.twctl = await TwCtl(cfg, proto, mode='w')
		await super().__init__(cfg, proto, minconf=minconf, quiet=quiet)

	async def create_data(self):
		in_data = self.twctl.mmid_ordered_dict
		block = self.twctl.rpc.get_block_from_minconf(self.minconf)
		for d in in_data:
			if d.type == 'mmgen':
				label = d.obj.sid
				if label not in self.data:
					self.data[label] = self.balance_info()
			else:
				label = 'Non-MMGen'

			amt = await self.twctl.get_balance(in_data[d]['addr'], block=block)

			self.data['TOTAL']['ge_minconf'] += amt
			self.data[label]['ge_minconf'] += amt

		del self.twctl

class EthereumTokenTwGetBalance(EthereumTwGetBalance):
	pass



#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
#
# Ported to Python 3 (added floor division) from ed25519ll package:
#   https://pypi.org/project/ed25519ll/
# This module is adapted from that package's pure-Python fallback
# implementation.  All functions and vars not required by MMGen have
# been removed.
#
# ed25519ll, a low-level ctypes wrapper for Ed25519 digital signatures by
# Daniel Holth <dholth@fastmail.fm> - http://bitbucket.org/dholth/ed25519ll/
#   This wrapper also contains a reasonably performant pure-Python fallback.
#   Unlike the reference implementation, the Python implementation does not
#   contain protection against timing attacks.
#
# Ed25519 digital signatures
# Based on http://ed25519.cr.yp.to/python/ed25519.py
# See also http://ed25519.cr.yp.to/software.html
# Adapted by Ron Garret
# Sped up considerably using coordinate transforms found on:
# http://www.hyperelliptic.org/EFD/g1p/auto-twisted-extended-1.html
# Specifically add-2008-hwcd-4 and dbl-2008-hwcd

q = 2**255 - 19

def expmod(b,e,m):
	if e == 0:
		return 1
	t = expmod(b,e//2,m)**2 % m
	if e & 1:
		t = (t*b) % m
	return t

# Can probably get some extra speedup here by replacing this with
# an extended-euclidean, but performance seems OK without that
def inv(x):
	return expmod(x,q-2,q)

# Faster (!) version based on:
# http://www.hyperelliptic.org/EFD/g1p/auto-twisted-extended-1.html

def xpt_add(pt1, pt2):
	(X1, Y1, Z1, T1) = pt1
	(X2, Y2, Z2, T2) = pt2
	A = ((Y1-X1)*(Y2+X2)) % q
	B = ((Y1+X1)*(Y2-X2)) % q
	C = (Z1*2*T2) % q
	D = (T1*2*Z2) % q
	E = (D+C) % q
	F = (B-A) % q
	G = (B+A) % q
	H = (D-C) % q
	X3 = (E*F) % q
	Y3 = (G*H) % q
	Z3 = (F*G) % q
	T3 = (E*H) % q
	return (X3, Y3, Z3, T3)

def xpt_double (pt):
	(X1, Y1, Z1, _) = pt
	A = (X1*X1)
	B = (Y1*Y1)
	C = (2*Z1*Z1)
	D = (-A) % q
	J = (X1+Y1) % q
	E = (J*J-A-B) % q
	G = (D+B) % q
	F = (G-C) % q
	H = (D-B) % q
	X3 = (E*F) % q
	Y3 = (G*H) % q
	Z3 = (F*G) % q
	T3 = (E*H) % q
	return (X3, Y3, Z3, T3)

def pt_xform (pt):
	(x, y) = pt
	return (x, y, 1, (x*y)%q)

def pt_unxform (pt):
	(x, y, z, _) = pt
	return ((x*inv(z))%q, (y*inv(z))%q)

def xpt_mult (pt, n):
	if n==0:
		return pt_xform((0,1))
	_ = xpt_double(xpt_mult(pt, n>>1))
	return xpt_add(_, pt) if n&1 else _

def scalarmult(pt, e):
	return pt_unxform(xpt_mult(pt_xform(pt), e))



#!/usr/bin/env python3

import os
from collections import namedtuple
from decimal import Decimal

from mmgen.cfg import Config
from mmgen.util import pp_fmt, ymsg
from mmgen.proto.btc.common import hash160
from mmgen.proto.cosmos.tx.protobuf import RawTx
from mmgen.proto.rune.tx.protobuf import (
	RuneTx,
	build_tx,
	build_swap_tx,
	tx_info,
	send_tx_parms,
	deposit_tx_parms,
	swap_tx_parms)

from ..include.common import vmsg, silence, end_silence

test_cfg = Config({'coin': 'rune', 'test_suite': True})

_pv = namedtuple('parse_vector', ['fn', 'txid', 'parms', 'null_fee'], defaults=[None])

parse_vectors = [
	_pv(
		'mainnet-tx-msgsend1.binpb',
		'36f91982c1911fe1aa66b44eed60e29175e5b8ae3301feef9158b7617779b00e',
		send_tx_parms(
			'thor1t60f02r8jvzjrhtnjgfj4ne6rs5wjnejwmj7fh',
			'thor166n4w5039meulfa3p6ydg60ve6ueac7tlt0jws',
			'12613.15290000',
			8000000,
			45060,
			302033,
			pubkey = '02f9cbb8409443ccf043f26d8f91c2550d2578ecc49bb3ad89d4e21a7882bf1e23',
			signature = 'd44b2e0c7546c5fae24a2c829757f49cce1bb29553f7e1a2f87c1ac2f1c46e22' # r
			'509d765fc605d85e8967639864622ebc7c39a1a93fc20cf0fe5d703c4aa3636d')), # s
	_pv(
		'mainnet-tx-msgdeposit1.binpb',
		'1089bbd54746bbc6a40e264d3ce8085561978739094c9c5aac59c569b4c28ba9',
		deposit_tx_parms(
			'THOR', 'RUNE', 'RUNE',
			'thor1lukwlve7hayy66qrdkp4k7sh0emjqwergy7tl3',
			'299.23861844',
			600000000,
			125632,
			348388,
			decimals = 8,
			memo = '=:LTC~LTC:thor1lukwlve7hayy66qrdkp4k7sh0emjqwergy7tl3:605926421/0/1',
			pubkey = '03da157f891abfe7822efb91f59667aa6cc6c3768a7e280caeb9ae243c969eb3e7',
			signature = '869399bcc2ccb9c9c286bdf214439ad132221cb8206547ceb012e06efbc3ff3e' # r
			'0ecc1ba4106702fb5b60cd7a8b94193ea71af5e6e860d978c49a6a63d97e4ded')), # s
	_pv(
		'mainnet-tx-msgdeposit2.binpb',
		'44f45b91e97558e63a11758ac3c186196b9b46f6331f32eff1256888ea879b62',
		deposit_tx_parms(
			'ETH', 'USDT-0XDAC17F958D2EE523A2206206994597C13D831EC7', 'USDT',
			'thor1xxncvuptvmgcl5ep7rry3xehtw97jsg9uyv6rn',
			'500.00000000',
			50000000,
			88176,
			104625,
			decimals = None,
			synth = False,
			trade = True,
			memo = '=:AVAX~AVAX:thor1xxncvuptvmgcl5ep7rry3xehtw97jsg9uyv6rn:2113883178',
			pubkey = '02a6e97e3f20809511500d8895117d4344badda9e6af4216d41b10a105d1070254',
			signature = 'b0673ab89781199d35b94051f26db30996f055abd71804d67fe4bdf33934bdb9' # r
			'30d830675d398c6d042328ef681bf1a23e856ed0ab33d948ea149489400db953'), # s
		null_fee = True),
	_pv(
		'mainnet-tx-msgdeposit3.binpb',
		'02d2fb2f2e5ac00ad4a31c37ffc43b72963f93598f8b3c8f4d3932c2e950b459',
		deposit_tx_parms(
			'ETH', 'USDC-0XA0B86991C6218B36C1D19D4A2E9EB0CE3606EB48', 'USDC',
			'thor1lukwlve7hayy66qrdkp4k7sh0emjqwergy7tl3',
			'2425.75973697',
			600000000,
			125632,
			375988,
			decimals = 8,
			synth = None,
			trade = True,
			memo = '=:THOR.RUNE:thor1lukwlve7hayy66qrdkp4k7sh0emjqwergy7tl3:127025748855/0/1',
			pubkey = '03da157f891abfe7822efb91f59667aa6cc6c3768a7e280caeb9ae243c969eb3e7',
			signature = 'be8324f6a1535b971d63715532e2e42ee3c35c05b81ed9b6bccd9a1765688eca'
			'4ca00a66119b07c5168c6c78e22299becc82d5f311ee79ae9183776b0dff3269'))
]

_bv = namedtuple('build_vector', ['txid', 'txid2', 'parms', 'null_fee'], defaults=[None])

build_vectors = [
	_bv(
		'3939612d0ddc55fd4d1c6ef118d1b2085a6655cb57d55ac9efd658467e039e0c',
		'e783ced14909a9e2a21c99b6b3b66fb38ba3bdf985876bb8d6cea02813d603a9',
		send_tx_parms(
			'thor1tx3nm6xfynq3re5ehtm6530z0pah9qjeu0r9nd',
			'thor1j5u6vlr8kzt76fe7896hsmurkhgn68j0z4qa6w',
			'123.456789',
			8000000,
			12345,
			37,
			wifkey = 'L5nWojqqMLq7wh3CfhxUNYQ38acABD6sUao9dfb8i5B5wSefCJXe')),
	_bv(
		'0d41e0ee40cd18a991cd8f0ef0e60e4c5bea898c53d54e00b6dddc0c9ce7edb7',
		'444e026fe5d0988da602dc22f0ff6172c080f1d2e2d66012d86f4afd314b78d6',
		deposit_tx_parms(
			'THOR', 'RUNE', 'RUNE',
			'thor18ug6p4zs5dsy0m3u69gf5md5ssdg8hqkk8aya4',
			'123.456789',
			8000000,
			12345,
			37,
			decimals = 8,
			memo = '=:MEMO',
			wifkey = 'Ky9bSjPUD35uUaY3JReXiESivnfxV6rLMsW1wTFyvVZwYXpX95vF'))
]

swap_build_vectors = [
	_bv(
		'0d41e0ee40cd18a991cd8f0ef0e60e4c5bea898c53d54e00b6dddc0c9ce7edb7',
		'444e026fe5d0988da602dc22f0ff6172c080f1d2e2d66012d86f4afd314b78d6',
		swap_tx_parms(
			'thor18ug6p4zs5dsy0m3u69gf5md5ssdg8hqkk8aya4',
			'123.456789',
			8000000,
			12345,
			37,
			memo = '=:MEMO',
			wifkey = 'Ky9bSjPUD35uUaY3JReXiESivnfxV6rLMsW1wTFyvVZwYXpX95vF'))
]

def test_tx(src, cfg, vec):

	proto = cfg._proto
	parms = vec.parms._replace(amt=Decimal(vec.parms.amt))
	if parms.pubkey:
		parms = parms._replace(
			pubkey = bytes.fromhex(parms.pubkey),
			signature = bytes.fromhex(parms.signature))

	assert src in ('parse', 'build', 'swapbuild')

	if src == 'parse':
		tx_in = open(os.path.join('test/ref/thorchain', vec.fn), 'br').read()
		tx = RuneTx.loads(tx_in)
		if not parms.from_addr:
			ymsg(f'Warning: missing test vector data for {vec.fn}')
		assert bytes(tx) == tx_in
	elif src == 'build':
		tx = build_tx(cfg, proto, parms, null_fee=vec.null_fee)
	elif src == 'swapbuild':
		tx = build_swap_tx(cfg, proto, parms)

	vmsg(pp_fmt(tx))

	msg_type = 'MsgSend' if tx.body.messages[0].id == '/types.MsgSend' else 'MsgDeposit'

	vmsg('\n  TX info:\n    ' + '\n    '.join(tx_info(tx, proto)) + '\n')

	tx.verify_sig(proto, parms.account_number)

	pubkey = tx.authInfo.signerInfos[0].publicKey.key.data
	vec_txid2 = getattr(vec, 'txid2', None)
	assert hash160(pubkey) == getattr(
		tx.body.messages[0].body,
		'fromAddress' if msg_type == 'MsgSend' else 'signer')
	if tx.txid not in (vec.txid, vec_txid2):
		raise ValueError(f'{tx.txid} not in ({vec.txid}, {vec_txid2})')
	if tx.txid == vec_txid2:
		ymsg('\nWarning: non-standard TxID produced')

	if src == 'parse' and parms.from_addr:
		built_tx = build_tx(cfg, proto, parms, null_fee=vec.null_fee)
		addr_from_pubkey = proto.encode_addr_bech32x(hash160(pubkey))
		assert addr_from_pubkey == parms.from_addr
		assert bytes(built_tx) == tx_in
		raw_tx = RawTx(bytes(tx.body), bytes(tx.authInfo), tx.signatures)
		assert bytes(raw_tx) == bytes(RawTx.loads(tx_in))
		assert bytes(raw_tx) == bytes(tx.raw)
		assert tx_in == bytes(tx)

class unit_tests:

	def txparse(self, name, ut, desc='transaction parsing and signature verification'):
		for vec in parse_vectors:
			test_tx('parse', test_cfg, vec)
		return True

	def txbuild(self, name, ut, desc='transaction building and signing (MsgSend, MsgDeposit)'):
		for vec in build_vectors:
			test_tx('build', test_cfg, vec)
		return True

	def swaptxbuild(self, name, ut, desc='transaction building and signing (Swap TX)'):
		for vec in swap_build_vectors:
			test_tx('swapbuild', test_cfg, vec)
		return True

	def rpc(self, name, ut, desc='remote RPC operations'):
		import sys, asyncio
		from mmgen.rpc import rpc_init
		from ..cmdtest_d.httpd.thornode.rpc import ThornodeRPCServer

		silence()
		regtest_cfg = Config({'coin': 'rune', 'regtest': True, 'test_suite': True})
		end_silence()

		thornode_server = ThornodeRPCServer()
		thornode_server.start()

		addr = 'thor1lukwlve7hayy66qrdkp4k7sh0emjqwergy7tl3'
		txhash = 'abcdef01' * 8
		txbytes = open('test/ref/thorchain/mainnet-tx-msgsend1.binpb', 'rb').read()

		async def main():

			rpc = await rpc_init(regtest_cfg)

			res = rpc.get_account_info(addr)
			assert res['address'] == addr
			assert res['account_number']
			assert res['sequence']

			res = rpc.get_tx_info(txhash)
			assert res['hash'] == txhash.upper()

			res = rpc.tx_op(txbytes.hex(), op='check_tx')
			assert res['code'] == 0

		asyncio.run(main())
		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.rune.tx.completed: THORChain completed transaction class
"""

from ....tx import completed as TxBase

from ...vm.tx.completed import Completed as VmCompleted

from .base import Base

class Completed(VmCompleted, Base, TxBase.Completed):

	@property
	def total_gas(self):
		return self.txobj['gas']

	@property
	def fee(self):
		return self.proto.coin_amt(self.dfl_fee, from_unit='satoshi')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.tx.completed: Ethereum completed transaction class
"""

from ....tx import completed as TxBase

from ...vm.tx.completed import Completed as VmCompleted

from .base import Base, TokenBase

class Completed(VmCompleted, Base, TxBase.Completed):
	fn_fee_unit = 'Mwei'

	@property
	def total_gas(self):
		return self.txobj['startGas']

	@property
	def fee(self):
		return self.fee_gasPrice2abs(self.txobj['gasPrice'].toWei())

class TokenCompleted(TokenBase, Completed):

	@property
	def change(self):
		return self.sum_inputs() - self.send_amt

	@property
	def total_gas(self):
		return self.txobj['startGas'] + (self.txobj['router_gas'] if self.is_swap else 0)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.info: transaction info class
"""

import importlib

from ..cfg import gc
from ..color import red, green, cyan, orange, blue, yellow, magenta
from ..util import msg, msg_r, decode_timestamp, make_timestr
from ..util2 import format_elapsed_hr

class TxInfo:

	def __init__(self, cfg, tx):
		self.cfg = cfg
		self.tx = tx

	def format(self, *, terse=False, sort='addr'):

		tx = self.tx

		if tx.is_swap:
			sort = 'raw'

		if tx.proto.base_proto == 'Ethereum':
			blockcount = None
		else:
			try:
				blockcount = tx.rpc.blockcount
			except:
				blockcount = None

		def get_max_mmwid(io):
			sel_f = (
				(lambda o: len(o.mmid) + 2) if io == tx.inputs else  # 2 = len('()')
				(lambda o: len(o.mmid) + (2, 8)[bool(o.is_chg)]))    # 6 = len(' (chg)')
			return max(max([sel_f(o) for o in io if o.mmid] or [0]), len(nonmm_str))

		nonmm_str = f'(non-{gc.proj_name} address)'
		max_mmwid = max(get_max_mmwid(tx.inputs), get_max_mmwid(tx.outputs))

		def gen_view():
			yield (self.txinfo_hdr_fs_short if terse else self.txinfo_hdr_fs).format(
				hdr = cyan(('SWAP ' if tx.is_swap else '') + 'TRANSACTION DATA'),
				i = tx.txid.hl(),
				a = tx.send_amt.hl(),
				c = tx.dcoin,
				r = green('True') if tx.is_replaceable() else red('False'),
				s = green('True') if tx.signed else red('False'),
				l = (
					orange(self.strfmt_locktime(terse=True)) if tx.locktime else
					green('None')))

			for attr, label in [('timestamp', 'Created:'), ('sent_timestamp', 'Sent:')]:
				if (val := getattr(tx, attr)) is not None:
					_ = decode_timestamp(val)
					yield f'  {label:8} {make_timestr(_)} ({format_elapsed_hr(_)})\n'

			if tx.chain != 'mainnet': # if mainnet has a coin-specific name, display it
				yield green(f'  Chain: {tx.chain.upper()}') + '\n'

			if tx.coin_txid:
				yield f'  {tx.coin} TxID: {tx.coin_txid.hl()}\n'

			if tx.is_swap:
				from ..swap.proto.thorchain import Memo, name
				data = tx.swap_memo.encode() if tx.proto.is_vm else tx.data_output.data
				if Memo.is_partial_memo(data):
					recv_mmid = getattr(tx, 'swap_recv_addr_mmid', None)
					p = Memo.parse(data.decode('ascii'))
					yield '  {} {}\n'.format(magenta('DEX Protocol:'), blue(name))
					yield '    Swap: {}\n'.format(orange(f'{tx.send_asset.name} => {tx.recv_asset.name}'))
					yield '    Dest: {}{}\n'.format(
						cyan(p.address),
						orange(f' {recv_mmid}') if recv_mmid else '')
					if not recv_mmid:
						yield yellow('    Warning: swap destination address is not a wallet address!\n')

			enl = ('\n', '')[bool(terse)]
			yield enl

			if tx.comment:
				yield f'  Comment: {tx.comment.hl()}\n{enl}'

			yield self.format_body(
				blockcount,
				nonmm_str,
				max_mmwid,
				enl,
				terse = terse,
				sort  = sort)

			iwidth = len(str(int(tx.sum_inputs())))

			yield self.txinfo_ftr_fs.format(
				i = tx.sum_inputs().fmt(iwidth, color=True),
				o = tx.sum_outputs().fmt(iwidth, color=True),
				C = tx.change.fmt(iwidth, color=True),
				s = tx.send_amt.fmt(iwidth, color=True),
				a = self.format_abs_fee(iwidth, color=True),
				r = self.format_rel_fee(),
				d = tx.dcoin,
				c = tx.coin)

			if tx.cfg.verbose:
				yield self.format_verbose_footer()

		return ''.join(gen_view())

	def view_with_prompt(self, prompt, *, pause=True):
		prompt += ' (y)es, (N)o, pager (v)iew, (t)erse view: '
		from ..term import get_char
		while True:
			reply = get_char(prompt, immed_chars='YyNnVvTt').strip('\n\r')
			msg('')
			if reply == '' or reply in 'Nn':
				break
			if reply in 'YyVvTt':
				self.view(
					pager = reply in 'Vv',
					pause = pause,
					terse = reply in 'Tt')
				break
			msg('Invalid reply')

	def view(self, *, pager=False, pause=True, terse=False):
		o = self.format(terse=terse)
		if pager:
			from ..ui import do_pager
			do_pager(o)
		else:
			msg('')
			msg_r(o)
			from ..term import get_char
			if pause:
				get_char('Press any key to continue: ')
				msg('')

def init_info(cfg, tx):
	return getattr(
		importlib.import_module(f'mmgen.proto.{tx.proto.base_proto_coin.lower()}.tx.info'),
		('Token' if tx.proto.tokensym else '') + 'TxInfo')(cfg, tx)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.new: Monero wallet ops for the MMGen Suite
"""

from ...color import red, pink
from ...util import msg, ymsg, make_timestr

from ...ui import keypress_confirm

from ..rpc import MoneroWalletRPC

from .spec import OpMixinSpec
from .wallet import OpWallet

class OpNew(OpMixinSpec, OpWallet):
	spec_id = 'newaddr_spec'
	spec_key = ((1, 'source'),)
	wallet_offline = True

	async def main(self):
		h = MoneroWalletRPC(self, self.source)
		h.open_wallet('Monero')

		desc = 'account' if self.account is None else 'address'
		label = (
			None if self.label == '' else
			'{} [{}]'.format(self.label or f'xmrwallet new {desc}', make_timestr()))

		wallet_data = h.get_wallet_data()

		if desc == 'address':
			h.print_acct_addrs(wallet_data, self.account)

		if keypress_confirm(
				self.cfg,
				'\nCreating new {a} for wallet {b}{c} with {d}\nOK?'.format(
					a = desc,
					b = red(str(self.source.idx)),
					c = '' if desc == 'account' else f', account {red("#"+str(self.account))}',
					d = 'label ' + pink('‘'+label+'’') if label else 'empty label')
				):

			if desc == 'address':
				h.create_new_addr(self.account, label=label)
			else:
				h.create_acct(label=label)

			wallet_data = h.get_wallet_data(print=desc=='account')

			if desc == 'address':
				h.print_acct_addrs(wallet_data, self.account)
		else:
			ymsg('\nOperation cancelled by user request')

		# wallet must be left open: otherwise the 'stop_wallet' RPC call used to stop the daemon will fail
		if self.cfg.no_stop_wallet_daemon:
			h.close_wallet('Monero')

		msg('')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>

"""
test.objattrtest_d.common: shared data for MMGen data objects tests
"""

from decimal import Decimal
from collections import namedtuple

from mmgen.obj import TwComment, CoinTxID, HexStr
from mmgen.seed import Seed, SeedID
from mmgen.subseed import SubSeedIdx, SubSeedList
from mmgen.seedsplit import SeedShareList, SeedShareCount, SeedSplitIDString, SeedShareIdx, MasterShareIdx
from mmgen.addr import AddrIdx

from ..include.common import cfg, getrand

atd = namedtuple('attrtest_entry', ['attrs', 'args', 'kwargs'])

seed_bin = getrand(32)

# use the constructors here! otherwise reassignment test might fail when
# reassignment would otherwise succeed
sample_objs = {
	'int':       int(1),
	'Decimal':   Decimal('0.01'),
	'NoneType':  None,
	'bool':      bool(True),
	'str':       str('foo'),
	'dict':      dict({'a': 1}),
	'list':      list([1]),
	'tuple':     tuple((1, 2)),
	'bytes':     bytes(1),

	'HexStr':    HexStr('ff'),
	'AddrIdx':   AddrIdx(1),
	'TwComment': TwComment('αω'),
	'CoinTxID':  CoinTxID('aa'*32),

	'SeedID':    SeedID(sid='F00F00BB'),
	'Seed':      Seed(cfg, seed_bin=seed_bin),

	'SubSeedList': SubSeedList(Seed(cfg, seed_bin=seed_bin)),
	'SubSeedIdx':  SubSeedIdx('1S'),

	'SeedSplitIDString': SeedSplitIDString('alice'),
	'SeedShareList':     SeedShareList(Seed(cfg, seed_bin=seed_bin), SeedShareCount(2)),
	'SeedShareIdx':      SeedShareIdx(1),
	'SeedShareCount':    SeedShareCount(2),
	'MasterShareIdx':    MasterShareIdx(7),
}



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.rpc: Monero wallet RPC methods for the MMGen Suite
"""

from collections import namedtuple

from ..obj import CoinTxID
from ..color import red, cyan, pink
from ..util import msg, msg_r, gmsg, gmsg_r, die
from ..addr import CoinAddr

from .include import gen_acct_addr_info, XMRWalletAddrSpec
from .file.tx import MoneroMMGenTX

class MoneroWalletRPC:

	def __init__(self, parent, d):
		self.parent = parent
		self.cfg = parent.cfg
		self.proto = parent.proto
		self.c = parent.c
		self.d = d
		self.fn = parent.get_wallet_fn(d)
		self.new_tx_cls = (
			MoneroMMGenTX.NewUnsigned if self.cfg.watch_only else
			MoneroMMGenTX.NewSigned)

	def open_wallet(self, desc=None, *, refresh=True):
		add_desc = desc + ' ' if desc else self.parent.add_wallet_desc
		gmsg_r(f'\n  Opening {add_desc}wallet...')
		self.c.call( # returns {}
			'open_wallet',
			filename = self.fn.name,
			password = self.d.wallet_passwd)
		gmsg('done')

		if refresh:
			gmsg_r(f'  Refreshing {add_desc}wallet...')
			ret = self.c.call('refresh')
			gmsg('done')
			if ret['received_money']:
				msg('  Wallet has received funds')

	def close_wallet(self, desc):
		gmsg_r(f'\n  Closing {desc} wallet...')
		self.c.call('close_wallet')
		gmsg_r('done')

	async def stop_wallet(self, desc):
		msg(f'Stopping {self.c.daemon.desc} on port {self.c.daemon.bind_port}')
		gmsg_r(f'\n  Stopping {desc} wallet...')
		await self.c.stop_daemon(quiet=True) # closes wallet
		gmsg_r('done')

	def gen_accts_info(self, accts_data, addrs_data, *, indent='    ', skip_empty_ok=False):
		from .ops import addr_width
		fs = indent + '  {I:<3} {A} {N} {B} {L}'
		yield indent + f'Accounts of wallet {self.fn.name}:'
		yield fs.format(
			I = '',
			A = 'Base Address'.ljust(addr_width),
			N = 'nAddrs',
			B = '  Unlocked Balance',
			L = 'Label')
		for i, e in enumerate(accts_data['subaddress_accounts']):
			if skip_empty_ok and self.cfg.skip_empty_accounts and not e['unlocked_balance']:
				continue
			ca = CoinAddr(self.proto, e['base_address'])
			from .ops import fmt_amt
			yield fs.format(
				I = str(e['account_index']),
				A = ca.hl(0) if self.cfg.full_address else ca.fmt(0, addr_width, color=True),
				N = red(str(len(addrs_data[i]['addresses'])).ljust(6)),
				B = fmt_amt(e['unlocked_balance']),
				L = pink(e['label']))

	def get_wallet_data(self, *, print=True, skip_empty_ok=False):
		accts_data = self.c.call('get_accounts')
		addrs_data = [
			self.c.call('get_address', account_index=i)
				for i in range(len(accts_data['subaddress_accounts']))
		]
		if print:
			msg('\n' + '\n'.join(self.gen_accts_info(accts_data, addrs_data, skip_empty_ok=skip_empty_ok)))
		bals_data = self.c.call('get_balance', all_accounts=True)
		return namedtuple('wallet_data', ['accts_data', 'addrs_data', 'bals_data'])(
			accts_data, addrs_data, bals_data)

	def create_acct(self, label=None):
		msg('\n    Creating new account...')
		ret = self.c.call('create_account', label=label)
		msg('      Index:   {}'.format(pink(str(ret['account_index']))))
		msg('      Address: {}'.format(cyan(ret['address'])))
		return (ret['account_index'], ret['address'])

	def get_last_acct(self, accts_data):
		msg('\n    Getting last account...')
		ret = accts_data['subaddress_accounts'][-1]
		msg('      Index:   {}'.format(pink(str(ret['account_index']))))
		msg('      Address: {}'.format(cyan(ret['base_address'])))
		return (ret['account_index'], ret['base_address'])

	def print_acct_addrs(self, wallet_data, account):
		msg('\n      Addresses of account #{} ({}):'.format(
			account,
			wallet_data.accts_data['subaddress_accounts'][account]['label']))
		msg('\n'.join(gen_acct_addr_info(self, wallet_data, account, indent='        ')))
		return wallet_data.addrs_data[account]['addresses']

	def create_new_addr(self, account, label):
		msg_r('\n    Creating new address: ')
		ret = self.c.call('create_address', account_index=account, label=label or '')
		msg(cyan(ret['address']))
		return ret['address']

	def get_last_addr(self, account, wallet_data, *, display=True):
		if display:
			msg('\n    Getting last address:')
		acct_addrs = wallet_data.addrs_data[account]['addresses']
		addr = acct_addrs[-1]['address']
		if display:
			msg('      ' + cyan(addr))
		return (addr, len(acct_addrs) - 1)

	def set_label(self, account, address_idx, label):
		return self.c.call(
			'label_address',
			index = { 'major': account, 'minor': address_idx },
			label = label
		)

	def make_transfer_tx(self, account, addr, amt):
		res = self.c.call(
			'transfer',
			account_index = account,
			destinations = [{
				'amount':  amt.to_unit('atomic'),
				'address': addr
			}],
			priority = self.cfg.priority or None,
			do_not_relay = True,
			get_tx_hex = True,
			get_tx_metadata = True
		)
		return self.new_tx_cls(
			cfg            = self.cfg,
			op             = self.parent.name,
			network        = self.proto.network,
			seed_id        = self.parent.kal.al_id.sid,
			source         = XMRWalletAddrSpec(self.parent.source.idx, self.parent.account, None),
			dest           = None,
			dest_address   = addr,
			txid           = res['tx_hash'],
			amount         = self.proto.coin_amt(res['amount'], from_unit='atomic'),
			fee            = self.proto.coin_amt(res['fee'], from_unit='atomic'),
			blob           = res['tx_blob'],
			metadata       = res['tx_metadata'],
			unsigned_txset = res['unsigned_txset'] if self.cfg.watch_only else None,
		)

	def make_sweep_tx(self, account, dest_acct, dest_addr_idx, addr, addrs_data):
		res = self.c.call(
			'sweep_all',
			address = addr,
			account_index = account,
			subaddr_indices = list(range(len(addrs_data[account]['addresses'])))
				if self.parent.name == 'sweep_all' else [],
			priority = self.cfg.priority or None,
			do_not_relay = True,
			get_tx_hex = True,
			get_tx_metadata = True
		)

		if len(res['tx_hash_list']) > 1:
			die(3, 'More than one TX required.  Cannot perform this sweep')

		return self.new_tx_cls(
			cfg            = self.cfg,
			op             = self.parent.name,
			network        = self.proto.network,
			seed_id        = self.parent.kal.al_id.sid,
			source         = XMRWalletAddrSpec(self.parent.source.idx, self.parent.account, None),
			dest           = XMRWalletAddrSpec(
								(self.parent.dest or self.parent.source).idx,
								dest_acct,
								dest_addr_idx),
			dest_address   = addr,
			txid           = res['tx_hash_list'][0],
			amount         = self.proto.coin_amt(res['amount_list'][0], from_unit='atomic'),
			fee            = self.proto.coin_amt(res['fee_list'][0], from_unit='atomic'),
			blob           = res['tx_blob_list'][0],
			metadata       = res['tx_metadata_list'][0],
			unsigned_txset = res['unsigned_txset'] if self.cfg.watch_only else None,
		)

	def relay_tx(self, tx_hex):
		ret = self.c.call('relay_tx', hex=tx_hex)
		try:
			msg('\n    Relayed {}'.format(CoinTxID(ret['tx_hash']).hl()))
		except:
			msg(f'\n   Server returned: {ret!s}')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.sign: Monero wallet ops for the MMGen Suite
"""

from ..file.tx import MoneroMMGenTX
from ..rpc import MoneroWalletRPC

from .wallet import OpWallet

class OpSign(OpWallet):
	action = 'signing transaction with'
	start_daemon = False

	async def main(self, fn, *, restart_daemon=True):
		if restart_daemon:
			await self.restart_wallet_daemon()
		tx = MoneroMMGenTX.Unsigned(self.cfg, fn)
		h = MoneroWalletRPC(self, self.addr_data[0])
		self.head_msg(tx.src_wallet_idx, h.fn)
		if restart_daemon:
			h.open_wallet(refresh=False)
		res = self.c.call(
			'sign_transfer',
			unsigned_txset = tx.data.unsigned_txset,
			export_raw = True,
			get_tx_keys = True
		)
		new_tx = MoneroMMGenTX.NewColdSigned(
			cfg            = self.cfg,
			txid           = res['tx_hash_list'][0],
			unsigned_txset = None,
			signed_txset   = res['signed_txset'],
			_in_tx         = tx,
		)
		return new_tx



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.export: Monero wallet ops for the MMGen Suite
"""

from ...util import gmsg, gmsg_r

from ..file.outputs import MoneroWalletOutputsFile
from ..rpc import MoneroWalletRPC

from .wallet import OpWallet

class OpExportOutputs(OpWallet):
	action = 'exporting outputs from'
	stem = 'process'
	sign = False

	async def process_wallet(self, d, fn, last):
		h = MoneroWalletRPC(self, d)
		h.open_wallet('source')

		if self.cfg.rescan_blockchain:
			gmsg_r('\n  Rescanning blockchain...')
			self.c.call('rescan_blockchain')
			gmsg('done')

		if self.cfg.rescan_spent:
			gmsg_r('\n  Rescanning spent outputs...')
			self.c.call('rescan_spent')
			gmsg('done')

		self.head_msg(d.idx, h.fn)
		for ftype in ('Unsigned', 'Signed'):
			old_fn = getattr(MoneroWalletOutputsFile, ftype).find_fn_from_wallet_fn(
				cfg             = self.cfg,
				wallet_fn       = fn,
				ret_on_no_match = True)
			if old_fn:
				old_fn.unlink()
		m = MoneroWalletOutputsFile.New(
			parent    = self,
			wallet_fn = fn,
			data      = self.c.call('export_outputs', all=True),
			sign      = self.sign,
		)
		m.write()
		return True

class OpExportOutputsSign(OpExportOutputs):
	opts = ('rescan_spent', 'rescan_blockchain')
	sign = True



#!/usr/bin/env python3

"""
test.modtest_d.gen: key/address generation unit tests for the MMGen suite
"""

from mmgen.color import blue
from mmgen.protocol import init_proto
from mmgen.key import PrivKey
from mmgen.addr import MMGenAddrType
from mmgen.addrgen import KeyGenerator, AddrGenerator
from mmgen.keygen import get_backends

from ..include.common import cfg, qmsg

# TODO: add viewkey checks
vectors = { # from tooltest2
	'btc': ((
		'5HwzecKMWD82ppJK3qMKpC7ohXXAwcyAN5VgdJ9PLFaAzpBG4sX',
		'1C5VPtgq9xQ6AcTgMAR3J6GDrs72HC4pS1',
		'legacy'
	), (
		'KwojSzt1VvW343mQfWQi3J537siAt5ktL2qbuCg1ZyKR8BLQ6UJm',
		'1Kz9fVSUMshzPejpzW9D95kScgA3rY6QxF',
		'compressed'
	), (
		'KwojSzt1VvW343mQfWQi3J537siAt5ktL2qbuCg1ZyKR8BLQ6UJm',
		'3AhjTiWHhVJAi1s5CfKMcLzYps12x3gZhg',
		'segwit'
	), (
		'KwojSzt1VvW343mQfWQi3J537siAt5ktL2qbuCg1ZyKR8BLQ6UJm',
		'bc1q6pqnfwwakuuejpm9w52ds342f9d5u36v0qnz7c',
		'bech32'),
	),
	'eth': ((
		'0000000000000000000000000000000000000000000000000000000000000001',
		'7e5f4552091a69125d5dfcb7b8c2659029395bdf',
		'ethereum',
		),),
	'xmr': ((
		'0000000000000000000000000000000000000000000000000000000000000001',
		'42nsXK8WbVGTNayQ6Kjw5UdgqbQY5KCCufdxdCgF7NgTfjC69Mna7DJSYyie77hZTQ8H92G2HwgFhgEUYnDzrnLnQdF28r3',
		'monero',
		),),
	'zec': ((
		'SKxny894fJe2rmZjeuoE6GVfNkWoXfPp8337VrLLNWG56FjqVUYR',
		'zceQDpyNwek7dKqF5ZuFGj7YrNVxh7X1aPkrVxDLVxWSiZAFDEuy5C7XNV8VhyZ3ghTPQ61xjCGiyLT3wqpiN1Yi6mdmaCq',
		'zcash_z',
		),),
}

def do_test(proto, wif, addr_chk, addr_type, internal_keccak):

	if internal_keccak:
		cfg.use_internal_keccak_module = True
		add_msg = ' (internal keccak module)'
	else:
		add_msg = ''

	at = MMGenAddrType(proto, addr_type)
	privkey = PrivKey(proto, wif=wif)

	for n, backend in enumerate(get_backends(at.pubkey_type)):

		kg = KeyGenerator(cfg, proto, at.pubkey_type, silent=n+1)
		qmsg(blue(f'  Testing backend {backend!r} for addr type {addr_type!r}{add_msg}'))

		data = kg.gen_data(privkey)

		for k, v in data._asdict().items():
			if v and k in ('pubkey', 'viewkey_bytes'):
				qmsg(f'    {k+":":19} {v.hex()}')

		ag = AddrGenerator( cfg, proto, addr_type)
		addr = ag.to_addr(data)
		qmsg(f'    addr:               {addr}\n')

		assert addr == addr_chk, f'{addr} != {addr_chk}'

	cfg.use_internal_keccak_module = False

def do_tests(coin, internal_keccak=False):
	proto = init_proto( cfg, coin)
	for wif, addr, addr_type in vectors[coin]:
		do_test(proto, wif, addr, addr_type, internal_keccak)
	return True

class unit_tests:

	altcoin_deps = ('eth', 'xmr', 'zec')

	def btc(self, name, ut):
		return do_tests('btc')

	def eth(self, name, ut):
		do_tests('eth')
		return do_tests('eth', internal_keccak=True)

	def xmr(self, name, ut):
		if not cfg.fast:
			do_tests('xmr')
		return do_tests('xmr', internal_keccak=True)

	def zec(self, name, ut):
		return do_tests('zec')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
platform.darwin.util: utilities for the macOS platform
"""

from pathlib import Path
from subprocess import run, PIPE, DEVNULL

from ...obj import MMGenLabel
from ...util import oneshot_warning

def get_device_size(path_or_label):
	import re
	cp = run(['diskutil', 'info', path_or_label], text=True, stdout=PIPE, check=True)
	res = [e for e in cp.stdout.splitlines() if 'Disk Size' in e]
	errmsg = '‘diskutil info’ output could not be parsed for device size'
	assert len(res) == 1, f'{errmsg}:\n{cp.stdout}'
	m = re.search(r'\((\d+) Bytes\)', res[0])
	assert m, f'{errmsg}:\n{res[0]}'
	return int(m[1])

class warn_ramdisk_too_small(oneshot_warning):
	message = 'requested ramdisk size ({}MB) too small, using {}MB instead'
	color = 'yellow'
	def __init__(self, usr_size, min_size):
		oneshot_warning.__init__(self, div=usr_size, fmt_args=[usr_size, min_size])

class RamDiskLabel(MMGenLabel):
	max_len = 24
	desc = 'ramdisk label'

class MacOSRamDisk:

	desc = 'ramdisk'
	min_size = 10 # 10MB is the minimum supported by hdiutil

	def __init__(self, cfg, label, size, *, path=None):
		if size < self.min_size:
			warn_ramdisk_too_small(size, self.min_size)
			size = self.min_size
		self.cfg = cfg
		self.label = RamDiskLabel(label)
		self.size = size # size in MiB
		self.dfl_path = Path('/Volumes') / self.label
		self.path = Path(path) if path else self.dfl_path

	def exists(self):
		return self.path.is_mount()

	def get_diskutil_size(self):
		return get_device_size(self.label) // (2**20)

	def create(self, *, quiet=False):
		redir = DEVNULL if quiet else None
		if self.exists():
			diskutil_size = self.get_diskutil_size()
			if diskutil_size != self.size:
				self.cfg._util.qmsg(f'Existing ramdisk has incorrect size {diskutil_size}MB, deleting')
				self.destroy()
			else:
				self.cfg._util.qmsg(f'{self.desc.capitalize()} {self.label.hl()} at path {self.path} already exists')
				return
		self.cfg._util.qmsg(f'Creating {self.desc} {self.label.hl()} of size {self.size}MB')
		cp = run(['hdiutil', 'attach', '-nomount', f'ram://{2048 * self.size}'], stdout=PIPE, check=True)
		self.dev_name = cp.stdout.decode().strip()
		self.cfg._util.qmsg(f'Created {self.desc} {self.label.hl()} [{self.dev_name}]')
		run(['diskutil', 'eraseVolume', 'APFS', self.label, self.dev_name], stdout=redir, check=True)
		diskutil_size = self.get_diskutil_size()
		assert diskutil_size == self.size, 'Reported ramdisk size {diskutil_size}MB is incorrect!'
		if self.path != self.dfl_path:
			run(['diskutil', 'umount', self.label], stdout=redir, check=True)
			self.path.mkdir(parents=True, exist_ok=True)
			run(['diskutil', 'mount', '-mountPoint', str(self.path.absolute()), self.label], stdout=redir, check=True)

	def destroy(self, *, quiet=False):
		if not self.exists():
			self.cfg._util.qmsg(f'{self.desc.capitalize()} {self.label.hl()} at path {self.path} not found')
			return
		redir = DEVNULL if quiet else None
		run(['diskutil', 'eject', self.label], stdout=redir, check=True)
		if not quiet:
			self.cfg._util.qmsg(f'Destroyed {self.desc} {self.label.hl()} at {self.path}')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tx.bump: Bitcoin transaction bump class
"""

from ....tx import bump as TxBase
from ....util import msg
from .new_swap import NewSwap
from .completed import Completed
from .unsigned import AutomountUnsigned

class Bump(Completed, NewSwap, TxBase.Bump):
	desc = 'fee-bumped transaction'

	def get_orig_rel_fee(self):
		return self.fee_abs2rel(self.sum_inputs() - self.sum_outputs())

	@property
	def min_fee(self):
		return self.sum_inputs() - self.sum_outputs() + self.relay_fee

	def bump_fee(self, idx, fee):
		self.update_output_amt(
			idx,
			self.sum_inputs() - self.sum_outputs(exclude=idx) - fee
		)

	def convert_and_check_fee(self, fee, desc):
		ret = super().convert_and_check_fee(fee, desc)
		if ret is False or self.new_outputs:
			return ret
		if ret < self.min_fee:
			msg('{} {c}: {} fee too small. Minimum fee: {} {c} ({} {})'.format(
				ret.hl(),
				desc,
				self.min_fee,
				self.fee_abs2rel(self.min_fee),
				self.rel_fee_desc,
				c = self.coin))
			return False
		output_amt = self.outputs[self.bump_output_idx].amt
		if ret >= output_amt:
			msg('{} {c}: {} fee too large. Maximum fee: <{} {c}'.format(
				ret.hl(),
				desc,
				output_amt.hl(),
				c = self.coin))
			return False
		return ret

class AutomountBump(Bump):
	desc      = 'unsigned fee-bumped automount transaction'
	ext       = AutomountUnsigned.ext
	automount = AutomountUnsigned.automount



#!/usr/bin/env python3

"""
test.modtest_d.bip39: BIP39 unit test for the MMGen suite
"""

from mmgen.color import blue, orange, purple
from mmgen.bip39 import bip39

from ..include.common import cfg, qmsg, vmsg

class unit_tests:

	vectors = (
		(
			"00000000000000000000000000000000",
			"abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon about"
		), (
			"7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f",
			"legal winner thank year wave sausage worth useful legal winner thank yellow"
		), (
			"80808080808080808080808080808080",
			"letter advice cage absurd amount doctor acoustic avoid letter advice cage above"
		), (
			"ffffffffffffffffffffffffffffffff",
			"zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo wrong"
		), (
			"000000000000000000000000000000000000000000000000",
			"abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon "
			"abandon abandon abandon abandon abandon agent"
		), (
			"7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f",
			"legal winner thank year wave sausage worth useful legal winner thank year wave sausage worth "
			"useful legal will"
		), (
			"808080808080808080808080808080808080808080808080",
			"letter advice cage absurd amount doctor acoustic avoid letter advice cage absurd amount doctor "
			"acoustic avoid letter always"
		), (
			"ffffffffffffffffffffffffffffffffffffffffffffffff",
			"zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo when"
		), (
			"0000000000000000000000000000000000000000000000000000000000000000",
			"abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon "
			"abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon art"
		), (
			"7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f7f",
			"legal winner thank year wave sausage worth useful legal winner thank year wave sausage worth "
			"useful legal winner thank year wave sausage worth title"
		), (
			"8080808080808080808080808080808080808080808080808080808080808080",
			"letter advice cage absurd amount doctor acoustic avoid letter advice cage absurd amount doctor "
			"acoustic avoid letter advice cage absurd amount doctor acoustic bless"
		), (
			"ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff",
			"zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo vote"
		), (
			"9e885d952ad362caeb4efe34a8e91bd2",
			"ozone drill grab fiber curtain grace pudding thank cruise elder eight picnic"
		), (
			"6610b25967cdcca9d59875f5cb50b0ea75433311869e930b",
			"gravity machine north sort system female filter attitude volume fold club stay feature office "
			"ecology stable narrow fog"
		), (
			"68a79eaca2324873eacc50cb9c6eca8cc68ea5d936f98787c60c7ebc74e6ce7c",
			"hamster diagram private dutch cause delay private meat slide toddler razor book happy fancy "
			"gospel tennis maple dilemma loan word shrug inflict delay length"
		), (
			"c0ba5a8e914111210f2bd131f3d5e08d",
			"scheme spot photo card baby mountain device kick cradle pact join borrow"
		), (
			"6d9be1ee6ebd27a258115aad99b7317b9c8d28b6d76431c3",
			"horn tenant knee talent sponsor spell gate clip pulse soap slush warm silver nephew swap uncle "
			"crack brave"
		), (
			"9f6a2878b2520799a44ef18bc7df394e7061a224d2c33cd015b157d746869863",
			"panda eyebrow bullet gorilla call smoke muffin taste mesh discover soft ostrich alcohol speed "
			"nation flash devote level hobby quick inner drive ghost inside"
		), (
			"23db8160a31d3e0dca3688ed941adbf3",
			"cat swing flag economy stadium alone churn speed unique patch report train"
		), (
			"8197a4a47f0425faeaa69deebc05ca29c0a5b5cc76ceacc0",
			"light rule cinnamon wrap drastic word pride squirrel upgrade then income fatal apart sustain "
			"crack supply proud access"
		), (
			"066dca1a2bb7e8a1db2832148ce9933eea0f3ac9548d793112d9a95c9407efad",
			"all hour make first leader extend hole alien behind guard gospel lava path output census museum "
			"junior mass reopen famous sing advance salt reform"
		), (
			"f30f8c1da665478f49b001d94c5fc452",
			"vessel ladder alter error federal sibling chat ability sun glass valve picture"
		), (
			"c10ec20dc3cd9f652c7fac2f1230f7a3c828389a14392f05",
			"scissors invite lock maple supreme raw rapid void congress muscle digital elegant little brisk "
			"hair mango congress clump"
		), (
			"f585c11aec520db57dd353c69554b21a89b20fb0650966fa0a9d6f74fd989d8f",
			"void come effort suffer camp survey warrior heavy shoot primary clutch crush open amazing screen "
			"patrol group space point ten exist slush involve unfold"
		)
	)

	def conversion(self, name, ut):

		vmsg('')
		qmsg(blue('Testing BIP39 conversion routines'))

		b = bip39()
		b.check_wordlist(cfg)

		vmsg('')
		qmsg('Checking seed to mnemonic conversion:')
		for v in self.vectors:
			chk = tuple(v[1].split())
			vmsg('    '+v[1])
			res = b.fromhex(v[0])
			assert res == chk, f'mismatch:\nres: {res}\nchk: {chk}'

		vmsg('')
		qmsg('Checking mnemonic to seed conversion:')
		for v in self.vectors:
			chk = v[0]
			vmsg('    '+chk)
			res = b.tohex(v[1].split())
			assert res == chk, f'mismatch:\nres: {res}\nchk: {chk}'

		qmsg('OK')
		vmsg('')

		return True

	def errors(self, name, ut):

		vmsg('')
		qmsg(blue('Testing error handling'))

		good_mn = "zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo wrong".split()
		bad_len_mn = "zoo zoo zoo".split()
		bad_chksum_mn = "zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo".split()
		bad_word_mn = "admire zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo".split()
		bad_seed = 'deadbeef'
		good_seed = 'deadbeef' * 4

		b = bip39()
		th = b.tohex
		fh = b.fromhex
		bad_data = (
			('hex',               'AssertionError', 'not a hexadecimal', lambda: fh('xx')),
			('seed len',          'AssertionError', 'invalid seed bit',  lambda: fh(bad_seed)),
			('mnemonic type',     'AssertionError', 'must be list',      lambda: th('string')),
			('arg (tostr=True)',  'AssertionError', "'tostr' must be",   lambda: fh(good_seed, tostr=True)),
			('pad len (fromhex)', 'AssertionError', "invalid 'pad' arg", lambda: fh(good_seed, pad=23)),
			('pad len (tohex)',   'AssertionError', "invalid 'pad' arg", lambda: th(good_mn, pad=23)),
			('word',              'MnemonicError',  "not in the BIP39",  lambda: th(bad_word_mn)),
			('checksum',          'MnemonicError',  "checksum",          lambda: th(bad_chksum_mn)),
			('seed phrase len',   'MnemonicError',  "phrase len",        lambda: th(bad_len_mn)),
		)

		ut.process_bad_data(bad_data)

		qmsg('OK')
		vmsg('')

		return True

	def genseed(self, name, ut):

		vmsg('')
		qmsg(blue('Testing seed generation with password'))
		fn = 'test/ref/98831F3A.bip39'

		vmsg(f'Seed source: {purple(fn)}')
		vmsg('KDF: PBKDF2-SHA512, 2048 rounds')
		with open(fn) as fh:
			mnemonic = fh.read().strip()

		seed_hex = bip39().generate_seed(mnemonic.split()).hex()
		vmsg(f'  Password: {orange("(null)")}\n    {seed_hex}')
		assert seed_hex == '3c30b98d3d9a713cf5a7a42f5dd27b3bf7f4d792d2b9225f6f519a0da978e13c6f36989ef2123b12a96d6ad5a443a95d61022ffaa9fbce8f946da7b67f75d339'

		passwd = 'passw0rd'
		seed_hex = bip39().generate_seed(mnemonic.split(), passwd=passwd).hex()
		vmsg(f'  Password: {orange(passwd)}\n    {seed_hex}')
		assert seed_hex == '7eb773bf60f1a5071f96736b6ddbe5c544a7b7740182a80493e29577e58b7cde011d4e38d26f65dab6c9fdebe5594e523447a1427ffd60746e6d04b4daa42eb1'

		qmsg('OK')
		vmsg('')

		return True



#!/usr/bin/env python3

"""
test.modtest_d.dep: dependency unit tests for the MMGen suite

  Test whether dependencies are installed and functional.
  No data verification is performed.
"""

import sys

from subprocess import run, PIPE

from mmgen.util import msg, rmsg, ymsg, gmsg
from mmgen.exception import NoLEDSupport

from ..include.common import cfg, vmsg, check_solc_ver

class unit_tests:

	altcoin_deps = ('solc', 'keccak', 'pysocks', 'semantic_version')
	win_skip = ('led', 'semantic_version')

	def secp256k1(self, name, ut):
		try:
			from mmgen.proto.secp256k1.secp256k1 import pubkey_gen
			pubkey_gen(bytes.fromhex('deadbeef'*8), 1)
			return True
		except ModuleNotFoundError as e:
			ymsg(f'{type(e).__name__}: {e}')
			msg('Installing secp256k1 module locally...')
			run(['python3', './setup.py', 'build_ext', '--inplace'], stdout=PIPE, stderr=PIPE, check=True)
			ymsg('The module has been installed.  Try re-running the test')
			sys.exit(1)
		return False

	def led(self, name, ut):
		from mmgen.led import LEDControl
		try:
			LEDControl(enabled=True)
		except NoLEDSupport:
			ymsg('Warning: no LED support on this platform')
		else:
			gmsg('LED support found!')
		return True

	def keccak(self, name, ut): # used by ETH, XMR
		from mmgen.util2 import get_keccak
		try:
			keccak_256 = get_keccak()
			keccak_256(b'abc')
		except Exception as e:
			rmsg(str(e))
			return False
		else:
			return True

	def pysocks(self, name, ut):
		import requests, urllib3
		urllib3.disable_warnings()
		session = requests.Session()
		session.trust_env = False
		session.proxies.update({'https':'socks5h://127.243.172.8:20677'})
		try:
			session.get('https://127.188.29.17', timeout=1)
		except Exception as e:
			if type(e).__name__ in ('ConnectionError', 'ConnectTimeout'):
				return True
			else:
				ymsg('{}: {}'.format(type(e).__name__, e))
				msg('Is the ‘pysocks’ package installed?')
		return False

	def cryptography(self, name, ut):
		from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
		from cryptography.hazmat.backends import default_backend
		c = Cipher(algorithms.AES(b'deadbeef'*4), modes.CTR(b'deadbeef'*2), backend=default_backend())
		encryptor = c.encryptor()
		encryptor.update(b'foo') + encryptor.finalize()
		return True

	def ecdsa(self, name, ut):
		import ecdsa
		pko = ecdsa.SigningKey.from_secret_exponent(12345678901234, curve=ecdsa.SECP256k1)
		pko.get_verifying_key().to_string().hex()
		return True

	def ripemd160(self, name, ut):
		import hashlib
		if hashlib.new.__name__ == 'hashlib_new_wrapper':
			ymsg('Warning: RIPEMD160 missing in hashlib, falling back on pure-Python implementation')
		hashlib.new('ripemd160')
		return True

	def gmpy(self, name, ut):
		from gmpy2 import context, set_context, sqrt, cbrt
		# context() parameters are platform-dependent!
		set_context(context(precision=75, round=1)) # OK for gmp 6.1.2 / gmpy 2.1.0
		return True

	def aiohttp(self, name, ut):
		import asyncio, aiohttp
		async def do():
			async with aiohttp.ClientSession(
				headers = {'Content-Type': 'application/json'},
				connector = aiohttp.TCPConnector(),
			):
				pass
		asyncio.run(do())
		return True

	def pexpect(self, name, ut):
		import pexpect
		from pexpect.popen_spawn import PopenSpawn
		return True

	def scrypt(self, name, ut):
		passwd, salt = b'foo', b'bar'
		N, r, p = 4, 8, 16
		buflen = 64

		vmsg('Testing builtin scrypt module (hashlib)')
		from hashlib import scrypt # max N == 14!!
		scrypt(password=passwd, salt=salt, n=2**N, r=r, p=p, maxmem=0, dklen=buflen)

		vmsg('Testing standalone scrypt module')
		import scrypt
		scrypt.hash(passwd, salt, N=2**N, r=r, p=p, buflen=buflen)

		return True

	def semantic_version(self, name, ut):
		from semantic_version import Version, NpmSpec
		return True

	def solc(self, name, ut):
		from mmgen.protocol import init_proto
		solc_ok = check_solc_ver()
		if solc_ok:
			cmd = [
				'python3',
				'scripts/create-token.py',
				'--coin=ETH',
				'--name=My Fake Token',
				'--symbol=FAKE',
				'--supply=100000000000000000000000000',
				'--decimals=18',
				'--stdout',
				init_proto(cfg, 'eth').checksummed_addr('deadbeef'*5),
			]
			cp = run(cmd, stdout=PIPE, stderr=PIPE)
			vmsg(cp.stderr.decode())
			if cp.returncode:
				msg(cp.stderr.decode())
				return False
		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.include.unit_test: Unit test framework for the MMGen suite
"""

import sys, os, time, importlib, platform, asyncio

from .test_init import repo_root

# for the unit tests, violate MMGen Project best practices and allow use of the dev tools
# in production code:
if not os.getenv('MMGEN_DEVTOOLS'):
	from mmgen.devinit import init_dev
	init_dev()

from mmgen.cfg import Config, gc, gv
from mmgen.color import gray, brown, orange, yellow, red
from mmgen.util import msg, msg_r, gmsg, ymsg, Msg

from test.include.common import set_globals, end_msg

def die(ev, s):
	msg((red if ev > 1 else yellow)(s))
	sys.exit(ev)

opts_data = {
	'text': {
		'desc': "Unit tests for the MMGen suite",
		'usage':'[options] [test | test.subtest]...',
		'options': """
-h, --help                Print this help message
-a, --no-altcoin-deps     Skip tests requiring altcoin daemons, libs or utils
-A, --no-daemon-autostart Don't start and stop daemons automatically
-D, --no-daemon-stop      Don't stop auto-started daemons after running tests
-f, --fast                Speed up execution by reducing rounds on some tests
-l, --list                List available tests
-L, --list-subtests       List available tests and subtests
-n, --names               Print command names instead of descriptions
-q, --quiet               Produce quieter output
-x, --exclude=T           Exclude tests 'T' (comma-separated)
-v, --verbose             Produce more verbose output
""",
	'notes': """
If no test is specified, all available tests are run
	"""
	}
}

if os.path.islink(Config.test_datadir):
	os.unlink(Config.test_datadir)

sys.argv.insert(1, '--skip-cfg-file')

cfg = Config(opts_data=opts_data)

type(cfg)._reset_ok += ('use_internal_keccak_module', 'debug_addrlist')

set_globals(cfg)

test_type = {
	'modtest.py':    'unit',
	'daemontest.py': 'daemon',
}[gc.prog_name]

test_subdir = gc.prog_name.removesuffix('.py') + '_d'

test_dir = os.path.join(repo_root, 'test', test_subdir)

all_tests = sorted(fn.removesuffix('.py') for fn in os.listdir(test_dir) if not fn.startswith('_'))

exclude = cfg.exclude.split(',') if cfg.exclude else []

if cfg.no_altcoin_deps:
	ymsg(f'{gc.prog_name}: skipping altcoin tests by user request')
	altcoin_tests = importlib.import_module(f'test.{test_subdir}').altcoin_tests

for e in exclude:
	if e not in all_tests:
		die(1, f'{e!r}: invalid parameter for --exclude (no such test)')

start_time = int(time.time())

if cfg.list:
	Msg(' '.join(all_tests))
	sys.exit(0)

if cfg.list_subtests:
	def gen():
		for test in all_tests:
			mod = importlib.import_module(f'test.{test_subdir}.{test}')
			if hasattr(mod, 'unit_tests'):
				t = getattr(mod, 'unit_tests')
				subtests = [k for k, v in t.__dict__.items() if type(v).__name__ == 'function' and k[0] != '_']
				yield fs.format(test, ' '.join(f'{subtest}' for subtest in subtests))
			else:
				yield test
	fs = '{:%s} {}' % max(len(t) for t in all_tests)
	Msg(fs.format('TEST', 'SUBTESTS') + '\n' + '\n'.join(gen()))
	sys.exit(0)

def silence():
	if not cfg.verbose:
		global stdout_save, stderr_save
		stdout_save = sys.stdout
		stderr_save = sys.stderr
		sys.stdout = sys.stderr = gv.stdout = gv.stderr = open(os.devnull, 'w')

def end_silence():
	if not cfg.verbose:
		global stdout_save, stderr_save
		sys.stdout = gv.stdout = stdout_save
		sys.stderr = gv.stderr = stderr_save

class UnitTestHelpers:

	def __init__(self, subtest_name):
		self.subtest_name = subtest_name

	def skip_msg(self, desc):
		cfg._util.qmsg(gray(
			f'Skipping {test_type} subtest {self.subtest_name.replace("_", "-")!r} for {desc}'
		))

	def process_bad_data(self, data, pfx='bad '):
		if os.getenv('PYTHONOPTIMIZE'):
			ymsg('PYTHONOPTIMIZE set, skipping error handling tests')
			return
		import re
		desc_w = max(len(e[0]) for e in data)
		exc_w = max(len(e[1]) for e in data)
		m_exc = '{!r}: incorrect exception type (expected {!r})'
		m_err = '{!r}: incorrect error msg (should match {!r}'
		m_noraise = "\nillegal action '{}{}' failed to raise an exception (expected {!r})"
		for (desc, exc_chk, emsg_chk, func) in data:
			try:
				cfg._util.vmsg_r('  {}{:{w}}'.format(pfx, desc+':', w=desc_w+1))
				ret = func()
				if type(ret).__name__ == 'coroutine':
					asyncio.run(ret)
			except Exception as e:
				exc = type(e).__name__
				emsg = e.args[0] if e.args else '(unspecified error)'
				cfg._util.vmsg(f' {exc:{exc_w}} [{emsg}]')
				assert exc == exc_chk, m_exc.format(exc, exc_chk)
				assert re.search(emsg_chk, emsg), m_err.format(emsg, emsg_chk)
			else:
				die(4, m_noraise.format(pfx, desc, exc_chk))

tests_seen = []

def run_test(test, subtest=None):

	def run_subtest(t, subtest):
		subtest_disp = subtest.replace('_', '-')
		msg(brown(f'Running {test_type} subtest ') + orange(f'{test}.{subtest_disp}'))

		if getattr(t, 'silence_output', False):
			silence()

		if hasattr(t, '_pre_subtest'):
			getattr(t, '_pre_subtest')(test, subtest, UnitTestHelpers(subtest))

		try:
			func = getattr(t, subtest.replace('-', '_'))
			c = func.__code__
			do_desc = c.co_varnames[c.co_argcount-1] == 'desc'
			if do_desc:
				if cfg.verbose:
					msg(f'Testing {func.__defaults__[0]}')
				elif not cfg.quiet:
					msg_r(f'Testing {func.__defaults__[0]}...')

			ret = func(test, UnitTestHelpers(subtest))
			if type(ret).__name__ == 'coroutine':
				ret = asyncio.run(ret)
			if do_desc and not cfg.quiet:
				msg('OK\n' if cfg.verbose else 'OK')
		except:
			if getattr(t, 'silence_output', False):
				end_silence()
			raise

		if hasattr(t, '_post_subtest'):
			getattr(t, '_post_subtest')(test, subtest, UnitTestHelpers(subtest))

		if getattr(t, 'silence_output', False):
			end_silence()

		if not ret:
			die(4, f'Unit subtest {subtest_disp!r} failed')

	if test not in tests_seen:
		gmsg(f'Running {test_type} test {test}')
		tests_seen.append(test)

	if cfg.no_altcoin_deps and test in altcoin_tests:
		msg(gray(f'Skipping {test_type} test {test!r} [--no-altcoin-deps]'))
		return

	mod = importlib.import_module(f'test.{test_subdir}.{test}')

	if hasattr(mod, 'unit_tests'): # new class-based API
		t = getattr(mod, 'unit_tests')()
		altcoin_deps = getattr(t, 'altcoin_deps', ())
		win_skip = getattr(t, 'win_skip', ())
		mac_skip = getattr(t, 'mac_skip', ())
		arm_skip = getattr(t, 'arm_skip', ())
		riscv_skip = getattr(t, 'riscv_skip', ())
		fast_skip = getattr(t, 'fast_skip', ())
		subtests = (
			[subtest] if subtest else
			[k for k, v in type(t).__dict__.items() if type(v).__name__ == 'function' and k[0] != '_']
		)
		if hasattr(t, '_pre'):
			t._pre()

		def subtest_skip_msg(name, add_msg):
			cfg._util.qmsg(gray(
				'Skipping {} subtest {!r} {}'.format(test_type, name.replace('_', '-'), add_msg)
			))

		for _subtest in subtests:
			if cfg.no_altcoin_deps and _subtest in altcoin_deps:
				subtest_skip_msg(_subtest, '[--no-altcoin-deps]')
				continue
			if cfg.fast and _subtest in fast_skip:
				subtest_skip_msg(_subtest, '[--fast]')
				continue
			if sys.platform == 'win32' and _subtest in win_skip:
				subtest_skip_msg(_subtest, 'for Windows platform')
				continue
			if sys.platform == 'darwin' and _subtest in mac_skip:
				subtest_skip_msg(_subtest, 'for macOS platform')
				continue
			if platform.machine() == 'aarch64' and _subtest in arm_skip:
				subtest_skip_msg(_subtest, 'for ARM platform')
				continue
			if platform.machine() == 'riscv64' and _subtest in riscv_skip:
				subtest_skip_msg(_subtest, 'for RISC-V platform')
				continue
			run_subtest(t, _subtest)
		if hasattr(t, '_post'):
			t._post()
	else:
		assert not subtest, f'{subtest!r}: subtests not supported for this {test_type} test'
		if not mod.unit_test().run_test(test, UnitTestHelpers(test)):
			die(4, 'Unit test {test!r} failed')

def main():
	for test in (cfg._args or all_tests):
		if '.' in test:
			test, subtest = test.split('.')
		else:
			subtest = None
		if test not in all_tests:
			die(1, f'{test!r}: test not recognized')
		if test not in exclude:
			run_test(test, subtest=subtest)
	end_msg(int(time.time()) - start_time)

from mmgen.main import launch
launch(func=main)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
fileutil: Routines that read, write, execute or stat files
"""

# 09 Mar 2024: make all utils accept Path instances as arguments

import sys, os

from .color import set_vt100
from .util import (
	msg,
	die,
	get_extension,
	is_utf8,
	capfirst,
	make_full_path,
	strip_comments,
)

def check_or_create_dir(path):
	try:
		os.listdir(path)
	except:
		if os.getenv('MMGEN_TEST_SUITE'):
			if os.path.exists(path): # path is a link or regular file
				from subprocess import run
				run(['rm', '-rf', str(path)])
				set_vt100()
		try:
			os.makedirs(path, 0o700)
		except:
			die(2, f'ERROR: unable to read or create path ‘{path}’')

def check_binary(args):
	from subprocess import run, DEVNULL
	try:
		run(args, stdout=DEVNULL, stderr=DEVNULL, check=True)
	except:
		die(2, f'{args[0]!r} binary missing, not in path, or not executable')
	set_vt100()

def shred_file(cfg, fn, *, iterations=30):
	check_binary(['shred', '--version'])
	from subprocess import run
	run(
		['shred', '--force', f'--iterations={iterations}', '--zero', '--remove=wipesync']
		+ (['--verbose'] if cfg.verbose else [])
		+ [str(fn)],
		check=True)
	set_vt100()

def _check_file_type_and_access(fname, ftype, *, blkdev_ok=False):

	import stat

	access, op_desc = (
		(os.W_OK, 'writ') if ftype in ('output file', 'output directory') else
		(os.R_OK, 'read'))

	if ftype == 'output directory':
		ok_types = [(stat.S_ISDIR, 'output directory')]
	else:
		ok_types = [
			(stat.S_ISREG, 'regular file'),
			(stat.S_ISLNK, 'symbolic link')
		]
		if blkdev_ok:
			if not sys.platform in ('win32',):
				ok_types.append((stat.S_ISBLK, 'block device'))

	try:
		mode = os.stat(fname).st_mode
	except:
		die('FileNotFound', f'Requested {ftype} ‘{fname}’ not found')

	for t in ok_types:
		if t[0](mode):
			break
	else:
		ok_list = ' or '.join(t[1] for t in ok_types)
		die(1, f'Requested {ftype} ‘{fname}’ is not a {ok_list}')

	if not os.access(fname, access):
		die(1, f'Requested {ftype} ‘{fname}’ is not {op_desc}able by you')

	return True

def check_infile(f, *, blkdev_ok=False):
	return _check_file_type_and_access(f, 'input file', blkdev_ok=blkdev_ok)

def check_outfile(f, *, blkdev_ok=False):
	return _check_file_type_and_access(f, 'output file', blkdev_ok=blkdev_ok)

def check_outfile_dir(fn, *, blkdev_ok=False):
	return _check_file_type_and_access(
		os.path.dirname(os.path.abspath(fn)), 'output directory', blkdev_ok=blkdev_ok)

def check_outdir(f):
	return _check_file_type_and_access(f, 'output directory')

def get_seed_file(cfg, *, nargs, wallets=None, invoked_as=None):

	wallets = wallets or cfg._args

	from .filename import find_file_in_dir
	from .wallet.mmgen import wallet

	wf = find_file_in_dir(wallet, cfg.data_dir)

	wd_from_opt = bool(cfg.hidden_incog_input_params or cfg.in_fmt) # have wallet data from opt?

	if len(wallets) + (wd_from_opt or bool(wf)) < nargs:
		if not wf:
			msg('No default wallet found, and no other seed source was specified')
		cfg._usage()
	elif len(wallets) > nargs:
		cfg._usage()
	elif len(wallets) == nargs and wf and invoked_as != 'gen':
		cfg._util.qmsg('Warning: overriding default wallet with user-supplied wallet')

	if wallets or wf:
		check_infile(wallets[0] if wallets else wf)

	return str(wallets[0]) if wallets else (wf, None)[wd_from_opt] # could be a Path instance

def _open_or_die(filename, mode, *, silent=False):
	try:
		return open(filename, mode)
	except:
		if silent:
			die(2, '')
		else:
			fn = {0:'STDIN', 1:'STDOUT', 2:'STDERR'}[filename] if isinstance(filename, int) else f'‘{filename}’'
			desc = 'reading' if 'r' in mode else 'writing'
			die(2, f'Unable to open file {fn} for {desc}')

def write_data_to_file(
		cfg,
		outfile,
		data,
		*,
		desc                  = 'data',
		ask_write             = False,
		ask_write_prompt      = '',
		ask_write_default_yes = True,
		ask_overwrite         = True,
		ask_tty               = True,
		no_tty                = False,
		no_stdout             = False,
		quiet                 = False,
		binary                = False,
		ignore_opt_outdir     = False,
		outdir                = None,
		check_data            = False,
		cmp_data              = None):

	outfile = str(outfile) # could be a Path instance

	if quiet:
		ask_tty = ask_overwrite = False

	if cfg.quiet:
		ask_overwrite = False

	if ask_write_default_yes is False or ask_write_prompt:
		ask_write = True

	def do_stdout():
		cfg._util.qmsg('Output to STDOUT requested')
		if cfg.stdin_tty:
			if no_tty:
				die(2, f'Printing {desc} to screen is not allowed')
			if (ask_tty and not cfg.quiet) or binary:
				from .ui import confirm_or_raise
				confirm_or_raise(
					cfg,
					message = '',
					action  = f'output {desc} to screen')
		else:
			try:
				of = os.readlink(f'/proc/{os.getpid()}/fd/1') # Linux
			except:
				of = None # Windows

			if of:
				if of[:5] == 'pipe:':
					if no_tty:
						die(2, f'Writing {desc} to pipe is not allowed')
					if ask_tty and not cfg.quiet:
						from .ui import confirm_or_raise
						confirm_or_raise(
							cfg,
							message = '',
							action  = f'output {desc} to pipe')
						msg('')
				of2, pd = os.path.relpath(of), os.path.pardir
				msg('Redirecting output to file {!r}'.format(of if of2[:len(pd)] == pd else of2))
			else:
				msg('Redirecting output to file')

		if binary and sys.platform == 'win32':
			import msvcrt
			msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)

		# MSWin workaround. See msg_r()
		try:
			sys.stdout.write(data.decode() if isinstance(data, bytes) else data)
		except:
			os.write(1, data if isinstance(data, bytes) else data.encode())

	def do_file(outfile, ask_write_prompt):
		if (outdir or (cfg.outdir and not ignore_opt_outdir)) and not os.path.isabs(outfile):
			outfile = make_full_path(str(outdir or cfg.outdir), outfile) # outdir could be Path instance

		if ask_write:
			if not ask_write_prompt:
				ask_write_prompt = f'Save {desc}?'
			from .ui import keypress_confirm
			if not keypress_confirm(
					cfg,
					ask_write_prompt,
					default_yes = ask_write_default_yes):
				die(1, f'{capfirst(desc)} not saved')

		hush = False
		if os.path.lexists(outfile) and ask_overwrite:
			from .ui import confirm_or_raise
			confirm_or_raise(
				cfg,
				message = '',
				action  = f'File {outfile!r} already exists\nOverwrite?')
			msg(f'Overwriting file {outfile!r}')
			hush = True

		# not atomic, but better than nothing
		# if cmp_data is empty, file can be either empty or non-existent
		if check_data:
			d = ''
			try:
				with open(outfile, ('r', 'rb')[bool(binary)]) as fp:
					d = fp.read()
			except:
				pass
			if d != cmp_data:
				die(3, f'{desc} in file {outfile!r} has been altered by some other program! Aborting file write')

		# To maintain portability, always open files in binary mode
		# If 'binary' option not set, encode/decode data before writing and after reading
		try:
			with _open_or_die(outfile, 'wb') as fp:
				fp.write(data if binary else data.encode())
		except:
			die(2, f'Failed to write {desc} to file {outfile!r}')

		if not (hush or quiet):
			msg(f'{capfirst(desc)} written to file {outfile!r}')

		return True

	if no_stdout:
		do_file(outfile, ask_write_prompt)
	elif cfg.stdout or outfile in ('', '-'):
		do_stdout()
	elif sys.stdin.isatty() and not sys.stdout.isatty():
		do_stdout()
	else:
		do_file(outfile, ask_write_prompt)

def get_words_from_file(cfg, infile, *, desc, quiet=False):

	if not quiet:
		cfg._util.qmsg(f'Getting {desc} from file ‘{infile}’')

	with _open_or_die(infile, 'rb') as fp:
		data = fp.read()

	try:
		words = data.decode().split()
	except:
		die(1, f'{capfirst(desc)} data must be UTF-8 encoded.')

	cfg._util.dmsg('Sanitized input: [{}]'.format(' '.join(words)))

	return words

def get_data_from_file(
		cfg,
		infile,
		*,
		desc   = 'data',
		dash   = False,
		silent = False,
		binary = False,
		quiet  = False):

	if not (cfg.quiet or silent or quiet):
		cfg._util.qmsg(f'Getting {desc} from file ‘{infile}’')

	with _open_or_die(
			(0 if dash and infile == '-' else infile),
			'rb',
			silent=silent) as fp:
		data = fp.read(cfg.max_input_size+1)

	if not binary:
		data = data.decode()

	if len(data) == cfg.max_input_size + 1:
		die('MaxInputSizeExceeded',
			f'Too much input data!  Max input data size: {cfg.max_input_size} bytes')

	return data

def get_lines_from_file(
		cfg,
		fn,
		*,
		desc          = 'data',
		trim_comments = False,
		quiet         = False,
		silent        = False):

	def decrypt_file_maybe():
		data = get_data_from_file(cfg, fn, desc=desc, binary=True, quiet=quiet, silent=silent)
		from .crypto import Crypto
		have_enc_ext = get_extension(fn) == Crypto.mmenc_ext
		if have_enc_ext or not is_utf8(data):
			m = ('Attempting to decrypt', 'Decrypting')[have_enc_ext]
			cfg._util.qmsg(f'{m} {desc} ‘{fn}’')
			data = Crypto(cfg).mmgen_decrypt_retry(data, desc=desc)
		return data

	lines = decrypt_file_maybe().decode().splitlines()
	if trim_comments:
		lines = strip_comments(lines)
	cfg._util.dmsg(f'Got {len(lines)} lines from file ‘{fn}’')
	return lines



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
addrgen: Address generation initialization code for the MMGen suite
"""

from .keygen import KeyGenerator # convenience import

# decorator for to_addr() and to_viewkey()
def check_data(orig_func):
	def f(self, data):
		assert data.pubkey_type == self.pubkey_type, 'addrgen.py:check_data() pubkey_type mismatch'
		assert data.compressed == self.compressed, (
	f'addrgen.py:check_data() expected compressed={self.compressed} but got compressed={data.compressed}'
		)
		return orig_func(self, data)
	return f

class addr_generator:

	class base:

		def __init__(self, cfg, proto, addr_type):
			self.proto = proto
			self.pubkey_type = addr_type.pubkey_type
			self.compressed = addr_type.compressed
			self.desc = f'AddrGenerator {type(self).__name__!r}'

	class keccak(base):

		def __init__(self, cfg, proto, addr_type):
			super().__init__(cfg, proto, addr_type)
			from .util2 import get_keccak
			self.keccak_256 = get_keccak(cfg)

def AddrGenerator(cfg, proto, addr_type):
	"""
	factory function returning an address generator for the specified address type
	"""

	package_map = {
		'legacy':     'btc',
		'compressed': 'btc',
		'segwit':     'btc',
		'bech32':     'btc',
		'bech32x':    'xchain',
		'monero':     'xmr',
		'ethereum':   'eth',
		'zcash_z':    'zec',
	}

	from .addr import MMGenAddrType

	if type(addr_type) is str:
		addr_type = MMGenAddrType(proto=proto, id_str=addr_type)
	elif type(addr_type) is MMGenAddrType:
		assert addr_type in proto.mmtypes, f'{addr_type}: invalid address type for coin {proto.coin}'
	else:
		raise TypeError(f"{type(addr_type)}: incorrect argument type for 'addr_type' arg")

	import importlib
	return getattr(
		importlib.import_module(f'mmgen.proto.{package_map[addr_type.name]}.addrgen'),
		addr_type.name)(cfg, proto, addr_type)
